"""Base class for encoders and generic multi encoders."""

import abc
import itertools

import tensorflow as tf

from opennmt.layers.reducer import JoinReducer


class Encoder(tf.keras.layers.Layer):
    """Base class for encoders."""

    def build_mask(self, inputs, sequence_length=None, dtype=tf.bool):
        """Builds a boolean mask for :obj:`inputs`."""
        if sequence_length is None:
            return None
        return tf.sequence_mask(
            sequence_length, maxlen=tf.shape(inputs)[1], dtype=dtype
        )

    @abc.abstractmethod
    def call(self, inputs, sequence_length=None, training=None):
        """Encodes an input sequence.

        Args:
          inputs: The inputs to encode of shape :math:`[B, T, ...]`.
          sequence_length: The length of each input with shape :math:`[B]`.
          training: Run in training mode.

        Returns:
          A tuple ``(outputs, state, sequence_length)``.
        """
        raise NotImplementedError()

    def __call__(self, inputs, sequence_length=None, **kwargs):
        """Encodes an input sequence.

        Args:
          inputs: A 3D ``tf.Tensor`` or ``tf.RaggedTensor``.
          sequence_length: A 1D ``tf.Tensor`` (optional if :obj:`inputs` is a
            ``tf.RaggedTensor``).
          training: Run the encoder in training mode.

        Returns:
          If :obj:`inputs` is a ``tf.Tensor``, the encoder returns a tuple
          ``(outputs, state, sequence_length)``. If :obj:`inputs` is a
          ``tf.RaggedTensor``, the encoder returns a tuple ``(outputs, state)``,
          where ``outputs`` is a ``tf.RaggedTensor``.
        """
        if isinstance(inputs, tf.RaggedTensor):
            is_ragged = True
            inputs, sequence_length = inputs.to_tensor(), inputs.row_lengths()
        else:
            is_ragged = False
        outputs, state, sequence_length = super().__call__(
            inputs, sequence_length=sequence_length, **kwargs
        )
        if is_ragged:
            outputs = tf.RaggedTensor.from_tensor(outputs, lengths=sequence_length)
            return outputs, state
        else:
            return outputs, state, sequence_length


class SequentialEncoder(Encoder):
    """An encoder that executes multiple encoders sequentially with optional
    transition layers.

    See for example "Cascaded Encoder" in https://arxiv.org/abs/1804.09849.
    """

    def __init__(
        self, encoders, states_reducer=JoinReducer(), transition_layer_fn=None
    ):
        """Initializes the parameters of the encoder.

        Args:
          encoders: A list of :class:`opennmt.encoders.Encoder`.
          states_reducer: A :class:`opennmt.layers.Reducer` to merge all
            states.
          transition_layer_fn: A callable or list of callables applied to the
            output of an encoder before passing it as input to the next. If it is a
            single callable, it is applied between every encoders. Otherwise, the
            ``i`` th callable will be applied between encoders ``i`` and ``i + 1``.

        Raises:
          ValueError: if :obj:`transition_layer_fn` is a list with a size not equal
            to the number of encoder transitions ``len(encoders) - 1``.
        """
        if (
            transition_layer_fn is not None
            and isinstance(transition_layer_fn, list)
            and len(transition_layer_fn) != len(encoders) - 1
        ):
            raise ValueError(
                "The number of transition layers must match the number of encoder "
                "transitions, expected %d layers but got %d."
                % (len(encoders) - 1, len(transition_layer_fn))
            )
        super().__init__()
        self.encoders = encoders
        self.states_reducer = states_reducer
        self.transition_layer_fn = transition_layer_fn

    def call(self, inputs, sequence_length=None, training=None):
        encoder_state = []

        for i, encoder in enumerate(self.encoders):
            if i > 0 and self.transition_layer_fn is not None:
                if isinstance(self.transition_layer_fn, list):
                    inputs = self.transition_layer_fn[i - 1](inputs)
                else:
                    inputs = self.transition_layer_fn(inputs)
            inputs, state, sequence_length = encoder(
                inputs, sequence_length=sequence_length, training=training
            )
            encoder_state.append(state)

        return (inputs, self.states_reducer(encoder_state), sequence_length)


class ParallelEncoder(Encoder):
    """An encoder that encodes its input with several encoders and reduces the
    outputs and states together. Additional layers can be applied on each encoder
    output and on the combined output (e.g. to layer normalize each encoder
    output).

    If the input is a single ``tf.Tensor``, the same input will be encoded by
    every encoders. Otherwise, when the input is a Python sequence (e.g. the non
    reduced output of a :class:`opennmt.inputters.ParallelInputter`),
    each encoder will encode its corresponding input in the sequence.

    See for example "Multi-Columnn Encoder" in https://arxiv.org/abs/1804.09849.
    """

    def __init__(
        self,
        encoders,
        outputs_reducer=None,
        states_reducer=None,
        outputs_layer_fn=None,
        combined_output_layer_fn=None,
    ):
        """Initializes the parameters of the encoder.

        Args:
          encoders: A list of :class:`opennmt.encoders.Encoder` or a single
            one, in which case the same encoder is applied to each input.
          outputs_reducer: A :class:`opennmt.layers.Reducer` to merge all
            outputs. If ``None``, defaults to
            :class:`opennmt.layers.JoinReducer`.
          states_reducer: A :class:`opennmt.layers.Reducer` to merge all
            states. If ``None``, defaults to
            :class:`opennmt.layers.JoinReducer`.
          outputs_layer_fn: A callable or list of callables applied to the
            encoders outputs If it is a single callable, it is applied on each
            encoder output. Otherwise, the ``i`` th callable is applied on
            encoder ``i`` output.
          combined_output_layer_fn: A callable to apply on the combined output
            (i.e. the output of :obj:`outputs_reducer`).

        Raises:
          ValueError: if :obj:`outputs_layer_fn` is a list with a size not equal
            to the number of encoders.
        """
        if (
            isinstance(encoders, list)
            and outputs_layer_fn is not None
            and isinstance(outputs_layer_fn, list)
            and len(outputs_layer_fn) != len(encoders)
        ):
            raise ValueError(
                "The number of output layers must match the number of encoders; "
                "expected %d layers but got %d."
                % (len(encoders), len(outputs_layer_fn))
            )
        super().__init__()
        self.encoders = encoders
        self.outputs_reducer = (
            outputs_reducer if outputs_reducer is not None else JoinReducer()
        )
        self.states_reducer = (
            states_reducer if states_reducer is not None else JoinReducer()
        )
        self.outputs_layer_fn = outputs_layer_fn
        self.combined_output_layer_fn = combined_output_layer_fn

    def call(self, inputs, sequence_length=None, training=None):
        all_outputs = []
        all_states = []
        all_sequence_lengths = []
        parallel_inputs = isinstance(inputs, (list, tuple))
        parallel_encoders = isinstance(self.encoders, (list, tuple))

        if parallel_encoders and parallel_inputs and len(inputs) != len(self.encoders):
            raise ValueError(
                "ParallelEncoder expects as many inputs as parallel encoders"
            )
        if parallel_encoders:
            encoders = self.encoders
        else:
            encoders = itertools.repeat(
                self.encoders, len(inputs) if parallel_inputs else 1
            )

        for i, encoder in enumerate(encoders):
            if parallel_inputs:
                encoder_inputs = inputs[i]
                length = sequence_length[i]
            else:
                encoder_inputs = inputs
                length = sequence_length

            outputs, state, length = encoder(
                encoder_inputs, sequence_length=length, training=training
            )

            if self.outputs_layer_fn is not None:
                if isinstance(self.outputs_layer_fn, list):
                    outputs = self.outputs_layer_fn[i](outputs)
                else:
                    outputs = self.outputs_layer_fn(outputs)

            all_outputs.append(outputs)
            all_states.append(state)
            all_sequence_lengths.append(length)

        outputs, sequence_length = self.outputs_reducer(
            all_outputs, sequence_length=all_sequence_lengths
        )

        if self.combined_output_layer_fn is not None:
            outputs = self.combined_output_layer_fn(outputs)

        return (outputs, self.states_reducer(all_states), sequence_length)



"""Define RNN-based decoders."""

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt.decoders import decoder
from opennmt.layers import bridge, common, rnn, transformer
from opennmt.layers.rnn import map_v1_weights_to_cell


class RNNDecoder(decoder.Decoder):
    """A basic RNN decoder."""

    def __init__(
        self,
        num_layers,
        num_units,
        bridge_class=None,
        cell_class=None,
        dropout=0.3,
        residual_connections=False,
        **kwargs
    ):
        """Initializes the decoder parameters.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer.
          bridge_class: A :class:`opennmt.layers.Bridge` class to pass the
            encoder state to the decoder. Default to
            :class:`opennmt.layers.ZeroBridge`.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell. Defaults to a LSTM cell.
          dropout: The probability to drop units in each layer output.
          residual_connections: If ``True``, each layer input will be added to its
            output.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.dropout = dropout
        self.cell = rnn.make_rnn_cell(
            num_layers,
            num_units,
            dropout=dropout,
            residual_connections=residual_connections,
            cell_class=cell_class,
        )
        if bridge_class is None:
            bridge_class = bridge.ZeroBridge
        self.bridge = bridge_class()

    def _get_initial_state(self, batch_size, dtype, initial_state=None):
        cell_initial_state = self.cell.get_initial_state(
            batch_size=batch_size, dtype=dtype
        )
        if initial_state is not None:
            cell_initial_state = self.bridge(initial_state, cell_initial_state)
        return cell_initial_state

    def step(
        self,
        inputs,
        timestep,
        state=None,
        memory=None,
        memory_sequence_length=None,
        training=None,
    ):
        outputs, state = self.cell(inputs, state, training=training)
        return outputs, state, None


class AttentionalRNNDecoder(RNNDecoder):
    """A RNN decoder with attention."""

    def __init__(
        self,
        num_layers,
        num_units,
        bridge_class=None,
        attention_mechanism_class=None,
        cell_class=None,
        dropout=0.3,
        residual_connections=False,
        first_layer_attention=False,
        attention_layer_activation=tf.math.tanh,
        **kwargs
    ):
        """Initializes the decoder parameters.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer.
          bridge: A :class:`opennmt.layers.Bridge` to pass the encoder state
            to the decoder.
          attention_mechanism_class: A class inheriting from
            ``tfa.seq2seq.AttentionMechanism``. Defaults to
            ``tfa.seq2seq.LuongAttention``.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell.
          dropout: The probability to drop units in each layer output.
          residual_connections: If ``True``, each layer input will be added to its
            output.
          first_layer_attention: If ``True``, output attention after the first layer.
          attention_layer_activation: The activation to produce the attentional hidden
            state. Defaults to tanh following Luong paper (equation (5) in
            https://arxiv.org/abs/1508.04025).
          **kwargs: Additional layer arguments.
        """
        super().__init__(
            num_layers,
            num_units,
            bridge_class=bridge_class,
            cell_class=cell_class,
            dropout=dropout,
            residual_connections=residual_connections,
            **kwargs,
        )
        if attention_mechanism_class is None:
            attention_mechanism_class = tfa.seq2seq.LuongAttention
        self.attention_mechanism = attention_mechanism_class(self.cell.output_size)

        def _add_attention(cell):
            # Produce Luong-style attentional hidden states.
            attention_layer = common.Dense(
                cell.output_size, use_bias=False, activation=attention_layer_activation
            )
            wrapper = tfa.seq2seq.AttentionWrapper(
                cell, self.attention_mechanism, attention_layer=attention_layer
            )
            return wrapper

        if first_layer_attention:
            self.cell.cells[0] = _add_attention(self.cell.cells[0])
        else:
            self.cell = _add_attention(self.cell)
        self.dropout = dropout
        self.first_layer_attention = first_layer_attention

    @property
    def support_alignment_history(self):
        return True

    def _get_initial_state(self, batch_size, dtype, initial_state=None):
        # Reset memory of attention mechanism.
        self.attention_mechanism.setup_memory(
            self.memory, memory_sequence_length=self.memory_sequence_length
        )
        decoder_state = self.cell.get_initial_state(batch_size=batch_size, dtype=dtype)
        if initial_state is not None:
            if self.first_layer_attention:
                cell_state = list(decoder_state)
                cell_state[0] = decoder_state[0].cell_state
                cell_state = self.bridge(initial_state, cell_state)
                cell_state[0] = decoder_state[0].clone(cell_state=cell_state[0])
                decoder_state = tuple(cell_state)
            else:
                cell_state = self.bridge(initial_state, decoder_state.cell_state)
                decoder_state = decoder_state.clone(cell_state=cell_state)
        return decoder_state

    def step(
        self,
        inputs,
        timestep,
        state=None,
        memory=None,
        memory_sequence_length=None,
        training=None,
    ):
        outputs, state = self.cell(inputs, state, training=training)
        outputs = common.dropout(outputs, self.dropout, training=training)
        if self.first_layer_attention:
            attention = state[0].alignments
        else:
            attention = state.alignments
        return outputs, state, attention

    def map_v1_weights(self, weights):
        if self.first_layer_attention or not isinstance(
            self.attention_mechanism, tfa.seq2seq.LuongAttention
        ):
            raise ValueError(
                "Can only map V1 weights for RNN decoder with Luong attention "
                "on the last layer"
            )
        m = super().map_v1_weights(weights)
        m += common.Dense.map_v1_weights(
            self.attention_mechanism.memory_layer, weights["memory_layer"]
        )
        weights = weights["decoder"]["attention_wrapper"]
        m += common.Dense.map_v1_weights(
            self.cell._attention_layers[0], weights["attention_layer"]
        )
        m += map_v1_weights_to_cell(self.cell._cell, weights)
        return m


class RNMTPlusDecoder(decoder.Decoder):
    """The RNMT+ decoder described in https://arxiv.org/abs/1804.09849."""

    def __init__(
        self, num_layers, num_units, num_heads, dropout=0.3, cell_class=None, **kwargs
    ):
        """Initializes the decoder parameters.

        Args:
          num_layers: The number of layers.
          num_units: The number of units in each layer.
          num_heads: The number of attention heads.
          dropout: The probability to drop units from the decoder input and in each
            layer output.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell. Defaults to a layer normalized LSTM cell.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        if cell_class is None:
            cell_class = tfa.rnn.LayerNormLSTMCell
        self.num_heads = num_heads
        self.num_units = num_units
        self.dropout = dropout
        self.cells = [cell_class(num_units) for _ in range(num_layers)]
        self.multi_head_attention = transformer.MultiHeadAttention(
            num_heads, num_units, dropout=dropout, return_attention=True
        )

    @property
    def support_alignment_history(self):
        return True

    def _get_initial_state(self, batch_size, dtype, initial_state=None):
        return tuple(
            cell.get_initial_state(batch_size=batch_size, dtype=dtype)
            for cell in self.cells
        )

    def step(
        self,
        inputs,
        timestep,
        state=None,
        memory=None,
        memory_sequence_length=None,
        training=None,
    ):
        inputs = common.dropout(inputs, rate=self.dropout, training=training)

        new_states = []
        last_outputs, state_0 = self.cells[0](inputs, state[0])
        new_states.append(state_0)

        if memory_sequence_length is not None:
            memory_mask = tf.sequence_mask(
                memory_sequence_length, maxlen=tf.shape(memory)[1]
            )
        else:
            memory_mask = None

        context, _, attention = self.multi_head_attention(
            tf.expand_dims(last_outputs, 1),
            memory=memory,
            mask=memory_mask,
            training=training,
        )
        attention = attention[:, 0, 0]  # Use the first head for the attention vector.
        context = tf.squeeze(context, axis=1)

        for i in range(1, len(self.cells)):
            inputs = tf.concat([last_outputs, context], axis=-1)
            outputs, state_i = self.cells[i](inputs, state[i], training=training)
            new_states.append(state_i)
            outputs = common.dropout(outputs, rate=self.dropout, training=training)
            if i >= 2:
                outputs += last_outputs
            last_outputs = outputs

        final = tf.concat([last_outputs, context], -1)
        return final, tuple(new_states), attention



"""Base class and functions for dynamic decoders."""

import abc

import tensorflow as tf

from opennmt import constants
from opennmt.inputters import text_inputter
from opennmt.layers import common
from opennmt.utils import decoding, misc


def get_sampling_probability(step, read_probability=None, schedule_type=None, k=None):
    """Returns the sampling probability as described in
    https://arxiv.org/abs/1506.03099.

    Args:
      step: The training step.
      read_probability: The probability to read from the inputs.
      schedule_type: The type of schedule: "constant", "linear", "exponential",
        or "inverse_sigmoid".
      k: The convergence constant.

    Returns:
      The probability to sample from the output ids as a 0D ``tf.Tensor`` or
      ``None`` if scheduled sampling is not configured.

    Raises:
      ValueError: if :obj:`schedule_type` is set but not :obj:`k` or if
       :obj:`schedule_type` is ``linear`` but an initial :obj:`read_probability`
       is not set.
      TypeError: if :obj:`schedule_type` is invalid.
    """
    if read_probability is None and schedule_type is None:
        return None

    if schedule_type is not None and schedule_type != "constant":
        if k is None:
            raise ValueError(
                "scheduled_sampling_k is required when scheduled_sampling_type is set"
            )

        step = tf.cast(step, tf.float32)
        k = tf.constant(k, tf.float32)

        if schedule_type == "linear":
            if read_probability is None:
                raise ValueError("Linear schedule requires an initial read probability")
            read_probability = min(read_probability, 1.0)
            read_probability = tf.maximum(read_probability - k * step, 0.0)
        elif schedule_type == "exponential":
            read_probability = tf.pow(k, step)
        elif schedule_type == "inverse_sigmoid":
            read_probability = k / (k + tf.exp(step / k))
        else:
            raise TypeError("Unknown scheduled sampling type: {}".format(schedule_type))

    return 1.0 - read_probability


class Decoder(tf.keras.layers.Layer):
    """Base class for decoders."""

    def __init__(self, num_sources=1, vocab_size=None, output_layer=None, **kwargs):
        """Initializes the decoder parameters.

        If you don't set one of :obj:`vocab_size` or :obj:`output_layer` here,
        you should later call the method :meth:`opennmt.decoders.Decoder.initialize`
        to initialize this decoder instance.

        Args:
          num_sources: The number of source contexts expected by this decoder.
          vocab_size: The output vocabulary size (optional if :obj:`output_layer` is set).
          output_layer: The output projection layer (optional).
          **kwargs: Additional layer arguments.

        Raises:
          ValueError: if the number of source contexts :obj:`num_sources` is not
            supported by this decoder.
        """
        if num_sources < self.minimum_sources or num_sources > self.maximum_sources:
            raise ValueError(
                "This decoder accepts between %d and %d source contexts, "
                "but received %d"
                % (self.minimum_sources, self.maximum_sources, num_sources)
            )
        super().__init__(**kwargs)
        self.num_sources = num_sources
        self.output_layer = None
        self.memory = None
        self.memory_sequence_length = None
        if vocab_size is not None or output_layer is not None:
            self.initialize(vocab_size=vocab_size, output_layer=output_layer)

    @property
    def minimum_sources(self):
        """The minimum number of source contexts supported by this decoder."""
        return 1

    @property
    def maximum_sources(self):
        """The maximum number of source contexts supported by this decoder."""
        return 1

    @property
    def support_alignment_history(self):
        """Returns ``True`` if this decoder can return the attention as alignment
        history."""
        return False

    @property
    def initialized(self):
        """Returns ``True`` if this decoder is initialized."""
        return self.output_layer is not None

    def initialize(self, vocab_size=None, output_layer=None):
        """Initializes the decoder configuration.

        Args:
          vocab_size: The target vocabulary size.
          output_layer: The output layer to use.

        Raises:
          ValueError: if both :obj:`vocab_size` and :obj:`output_layer` are not set.
        """
        if self.initialized:
            return
        if output_layer is not None:
            self.output_layer = output_layer
        else:
            if vocab_size is None:
                raise ValueError("One of vocab_size and output_layer must be set")
            self.output_layer = common.Dense(vocab_size)

    def reuse_embeddings(self, embeddings):
        """Reuses embeddings in the decoder output layer.

        Args:
          embeddings: The embeddings matrix to reuse.

        Raises:
          RuntimeError: if the decoder was not initialized.
        """
        self._assert_is_initialized()
        self.output_layer.set_kernel(embeddings, transpose=True)

    def initial_state(
        self,
        memory=None,
        memory_sequence_length=None,
        initial_state=None,
        batch_size=None,
        dtype=None,
    ):
        """Returns the initial decoder state.

        Args:
          memory: Memory values to query.
          memory_sequence_length: Memory values length.
          initial_state: An initial state to start from, e.g. the last encoder
            state.
          batch_size: The batch size to use.
          dtype: The dtype of the state.

        Returns:
          A nested structure of tensors representing the decoder state.

        Raises:
          RuntimeError: if the decoder was not initialized.
          ValueError: if one of :obj:`batch_size` or :obj:`dtype` is not set and
            neither :obj:`initial_state` nor :obj:`memory` are not passed.
          ValueError: if the number of source contexts (:obj:`memory`) does not
            match the number defined at the decoder initialization.
        """
        self._assert_is_initialized()
        self._assert_memory_is_compatible(memory, memory_sequence_length)
        self.memory = memory
        self.memory_sequence_length = memory_sequence_length
        if batch_size is None or dtype is None:
            sentinel = tf.nest.flatten(memory)[0]
            if sentinel is None:
                sentinel = tf.nest.flatten(initial_state)[0]
            if sentinel is None:
                raise ValueError(
                    "If batch_size or dtype are not set, then either "
                    "memory or initial_state should be set"
                )
            if batch_size is None:
                batch_size = tf.shape(sentinel)[0]
            if dtype is None:
                dtype = sentinel.dtype
        return self._get_initial_state(batch_size, dtype, initial_state=initial_state)

    def call(
        self,
        inputs,
        length_or_step=None,
        state=None,
        input_fn=None,
        sampling_probability=None,
        training=None,
    ):
        """Runs the decoder layer on either a complete sequence (e.g. for training
        or scoring), or a single timestep (e.g. for iterative decoding).

        Args:
          inputs: The inputs to decode, can be a 3D (training) or 2D (iterative
            decoding) tensor.
          length_or_step: For 3D :obj:`inputs`, the length of each sequence. For 2D
            :obj:`inputs`, the current decoding timestep.
          state: The decoder state.
          input_fn: A callable taking sampled ids and returning the decoding inputs.
          sampling_probability: When :obj:`inputs` is the full sequence, the
            probability to read from the last sample instead of the true target.
          training: Run in training mode.

        Returns:
          A tuple with the logits, the decoder state, and an attention vector.

        Raises:
          RuntimeError: if the decoder was not initialized.
          ValueError: if the :obj:`inputs` rank is different than 2 or 3.
          ValueError: if :obj:`length_or_step` is invalid.
        """
        self._assert_is_initialized()
        rank = inputs.shape.ndims
        if rank == 2:
            if length_or_step.shape.ndims != 0:
                raise ValueError(
                    "length_or_step should be a scalar with the current timestep"
                )
            outputs, state, attention = self.step(
                inputs,
                length_or_step,
                state=state,
                memory=self.memory,
                memory_sequence_length=self.memory_sequence_length,
                training=training,
            )
            logits = self.output_layer(outputs)
        elif rank == 3:
            if length_or_step.shape.ndims != 1:
                raise ValueError(
                    "length_or_step should contain the length of each sequence"
                )
            logits, state, attention = self.forward(
                inputs,
                sequence_length=length_or_step,
                initial_state=state,
                memory=self.memory,
                memory_sequence_length=self.memory_sequence_length,
                input_fn=input_fn,
                sampling_probability=sampling_probability,
                training=training,
            )
        else:
            raise ValueError("Unsupported input rank %d" % rank)
        return logits, state, attention

    def forward(
        self,
        inputs,
        sequence_length=None,
        initial_state=None,
        memory=None,
        memory_sequence_length=None,
        input_fn=None,
        sampling_probability=None,
        training=None,
    ):
        """Runs the decoder on full sequences.

        Args:
          inputs: The 3D decoder input.
          sequence_length: The length of each input sequence.
          initial_state: The initial decoder state.
          memory: Memory values to query.
          memory_sequence_length: Memory values length.
          input_fn: A callable taking sampled ids and returning the decoding inputs.
          sampling_probability: The probability to read from the last sample instead
            of the true target.
          training: Run in training mode.

        Returns:
          A tuple with the logits, the decoder state, and the attention
          vector.
        """
        _ = sequence_length
        fused_projection = True
        if sampling_probability is not None:
            if input_fn is None:
                raise ValueError(
                    "input_fn is required when a sampling probability is set"
                )
            if not tf.is_tensor(sampling_probability) and sampling_probability == 0:
                sampling_probability = None
            else:
                fused_projection = False
                tf.summary.scalar("sampling_probability", sampling_probability)

        batch_size, max_step, _ = misc.shape_list(inputs)
        inputs_ta = tf.TensorArray(inputs.dtype, size=max_step)
        inputs_ta = inputs_ta.unstack(tf.transpose(inputs, perm=[1, 0, 2]))

        def _maybe_sample(true_inputs, logits):
            # Read from samples with a probability.
            draw = tf.random.uniform([batch_size])
            read_sample = tf.less(draw, sampling_probability)
            sampled_ids = tf.random.categorical(logits, 1)
            sampled_inputs = input_fn(tf.squeeze(sampled_ids, 1))
            inputs = tf.where(
                tf.broadcast_to(tf.expand_dims(read_sample, -1), tf.shape(true_inputs)),
                x=sampled_inputs,
                y=true_inputs,
            )
            return inputs

        def _body(step, state, inputs, outputs_ta, attention_ta):
            outputs, state, attention = self.step(
                inputs,
                step,
                state=state,
                memory=memory,
                memory_sequence_length=memory_sequence_length,
                training=training,
            )
            next_inputs = tf.cond(
                step + 1 < max_step,
                true_fn=lambda: inputs_ta.read(step + 1),
                false_fn=lambda: tf.zeros_like(inputs),
            )
            if not fused_projection:
                outputs = self.output_layer(outputs)
            if sampling_probability is not None:
                next_inputs = _maybe_sample(next_inputs, outputs)
            outputs_ta = outputs_ta.write(step, outputs)
            if attention is not None:
                attention_ta = attention_ta.write(step, attention)
            return step + 1, state, next_inputs, outputs_ta, attention_ta

        step = tf.constant(0, dtype=tf.int32)
        outputs_ta = tf.TensorArray(inputs.dtype, size=max_step)
        attention_ta = tf.TensorArray(inputs.dtype, size=max_step)

        _, state, _, outputs_ta, attention_ta = tf.while_loop(
            lambda *arg: True,
            _body,
            loop_vars=(
                step,
                initial_state,
                inputs_ta.read(0),
                outputs_ta,
                attention_ta,
            ),
            parallel_iterations=32,
            swap_memory=True,
            maximum_iterations=max_step,
        )

        outputs = tf.transpose(outputs_ta.stack(), perm=[1, 0, 2])
        logits = self.output_layer(outputs) if fused_projection else outputs
        attention = None
        if self.support_alignment_history:
            attention = tf.transpose(attention_ta.stack(), perm=[1, 0, 2])
        return logits, state, attention

    @abc.abstractmethod
    def step(
        self,
        inputs,
        timestep,
        state=None,
        memory=None,
        memory_sequence_length=None,
        training=None,
    ):
        """Runs one decoding step.

        Args:
          inputs: The 2D decoder input.
          timestep: The current decoding step.
          state: The decoder state.
          memory: Memory values to query.
          memory_sequence_length: Memory values length.
          training: Run in training mode.

        Returns:
          A tuple with the decoder outputs, the decoder state, and the attention
          vector.
        """
        raise NotImplementedError()

    def dynamic_decode(
        self,
        embeddings,
        start_ids,
        end_id=constants.END_OF_SENTENCE_ID,
        initial_state=None,
        decoding_strategy=None,
        sampler=None,
        maximum_iterations=None,
        minimum_iterations=0,
        tflite_output_size=None,
    ):
        """Decodes dynamically from :obj:`start_ids`.

        Args:
          embeddings: Target embeddings or :class:`opennmt.inputters.WordEmbedder`
            to apply on decoded ids.
          start_ids: Initial input IDs of shape :math:`[B]`.
          end_id: ID of the end of sequence token.
          initial_state: Initial decoder state.
          decoding_strategy: A :class:`opennmt.utils.DecodingStrategy`
            instance that define the decoding logic. Defaults to a greedy search.
          sampler: A :class:`opennmt.utils.Sampler` instance that samples
            predictions from the model output. Defaults to an argmax sampling.
          maximum_iterations: The maximum number of iterations to decode for.
          minimum_iterations: The minimum number of iterations to decode for.
          tflite_output_size: If not None will run TFLite safe, is the size of 1D output tensor.

        Returns:
          A :class:`opennmt.utils.DecodingResult` instance.

        See Also:
          :func:`opennmt.utils.dynamic_decode`
        """
        if isinstance(embeddings, text_inputter.WordEmbedder):
            input_fn = lambda ids: embeddings({"ids": ids})
        else:
            input_fn = lambda ids: tf.nn.embedding_lookup(embeddings, ids)

        # TODO: find a better way to pass the state reorder flags.
        if hasattr(decoding_strategy, "_set_state_reorder_flags"):
            state_reorder_flags = self._get_state_reorder_flags()
            decoding_strategy._set_state_reorder_flags(state_reorder_flags)

        return decoding.dynamic_decode(
            lambda ids, step, state: self(input_fn(ids), step, state),
            start_ids,
            end_id=end_id,
            initial_state=initial_state,
            decoding_strategy=decoding_strategy,
            sampler=sampler,
            maximum_iterations=maximum_iterations,
            minimum_iterations=minimum_iterations,
            attention_history=self.support_alignment_history,
            attention_size=tf.shape(self.memory)[1]
            if self.support_alignment_history
            else None,
            tflite_output_size=tflite_output_size,
        )

    def map_v1_weights(self, weights):
        return self.output_layer.map_v1_weights(weights["dense"])

    @abc.abstractmethod
    def _get_initial_state(self, batch_size, dtype, initial_state=None):
        """Returns the decoder initial state.

        Args:
          batch_size: The batch size of the returned state.
          dtype; The data type of the state.
          initial_state: A state to start from.

        Returns:
          The decoder state as a nested structure of tensors.
        """
        raise NotImplementedError()

    def _get_state_reorder_flags(self):
        """Returns a structure that marks states that should be reordered during beam search.
        By default all states are reordered.

        Returns:
          The same structure as the decoder state with tensors replaced by booleans.
        """
        return None

    def _assert_is_initialized(self):
        """Raises an expection if the decoder was not initialized."""
        if not self.initialized:
            raise RuntimeError("The decoder was not initialized")

    def _assert_memory_is_compatible(self, memory, memory_sequence_length):
        """Raises an expection if the memory layout is not compatible with this decoder."""

        def _num_elements(obj):
            if obj is None:
                return 0
            elif isinstance(obj, (list, tuple)):
                return len(obj)
            else:
                return 1

        num_memory = _num_elements(memory)
        num_length = _num_elements(memory_sequence_length)
        if num_memory != num_length and memory_sequence_length is not None:
            raise ValueError(
                "got %d memory values but %d length vectors" % (num_memory, num_length)
            )
        if num_memory != self.num_sources:
            raise ValueError(
                "expected %d source contexts, but got %d"
                % (self.num_sources, num_memory)
            )



"""Module defining inputters.

Inputters implement the logic of transforming raw data to vectorized inputs,
e.g., from a line of text to a sequence of embeddings.
"""

from opennmt.inputters.inputter import (
    ExampleInputter,
    ExampleInputterAdapter,
    Inputter,
    MixedInputter,
    MultiInputter,
    ParallelInputter,
)
from opennmt.inputters.record_inputter import (
    SequenceRecordInputter,
    create_sequence_records,
    write_sequence_record,
)
from opennmt.inputters.text_inputter import (
    CharConvEmbedder,
    CharEmbedder,
    CharRNNEmbedder,
    TextInputter,
    WordEmbedder,
    add_sequence_controls,
    load_pretrained_embeddings,
)



"""Define word-based embedders."""

import abc
import collections
import os

import numpy as np
import tensorflow as tf

from google.protobuf import text_format
from tensorboard.plugins import projector

from opennmt import constants, tokenizers
from opennmt.data import dataset as dataset_util
from opennmt.data import text, vocab
from opennmt.inputters.inputter import Inputter
from opennmt.layers import common
from opennmt.utils import misc


def save_embeddings_metadata(
    log_dir, variable_name, vocabulary_file, num_oov_buckets=1
):
    """Registers an embedding variable for visualization in TensorBoard.

    This function registers :obj:`variable_name` in the ``projector_config.pbtxt``
    file and generates metadata from :obj:`vocabulary_file` to attach a label
    to each word ID.

    Args:
      log_dir: The active log directory.
      variable_name: The variable name in the checkpoint.
      vocabulary_file: The associated vocabulary file.
      num_oov_buckets: The number of additional unknown tokens.
    """
    # Assume it ends with /.ATTRIBUTES/VALUE
    filename = "%s.txt" % "_".join(variable_name.split("/")[:-2])
    metadata_path = os.path.join(log_dir, filename)

    with tf.io.gfile.GFile(vocabulary_file) as src, tf.io.gfile.GFile(
        metadata_path, mode="w"
    ) as dst:
        ws_index = 0
        for line in src:
            # The TensorBoard code checks line.trim().length == 0 when loading the
            # metadata file so make sure lines are not dropped.
            if not line.replace("\uFEFF", "").strip():
                dst.write("<whitespace%d>\n" % ws_index)
                ws_index += 1
            else:
                dst.write(line)
        if num_oov_buckets == 1:
            dst.write("<unk>\n")
        else:
            for i in range(num_oov_buckets):
                dst.write("<unk%d>\n" % i)

    config = projector.ProjectorConfig()

    # If the projector file exists, load it.
    config_path = os.path.join(log_dir, "projector_config.pbtxt")
    if tf.io.gfile.exists(config_path):
        with tf.io.gfile.GFile(config_path) as config_file:
            text_format.Merge(config_file.read(), config)

    # If this embedding is already registered, just update the metadata path.
    exists = False
    for meta in config.embeddings:
        if meta.tensor_name == variable_name:
            meta.metadata_path = filename
            exists = True
            break

    if not exists:
        embedding = config.embeddings.add()
        embedding.tensor_name = variable_name
        embedding.metadata_path = filename

    with tf.io.gfile.GFile(config_path, "w") as config_file:
        config_file.write(text_format.MessageToString(config))


def load_pretrained_embeddings(
    embedding_file,
    vocabulary_file,
    num_oov_buckets=1,
    with_header=True,
    case_insensitive_embeddings=True,
):
    """Returns pretrained embeddings relative to the vocabulary.

    The :obj:`embedding_file` must have the following format:

    .. code-block:: text

        N M
        word1 val1 val2 ... valM
        word2 val1 val2 ... valM
        ...
        wordN val1 val2 ... valM

    or if :obj:`with_header` is ``False``:

    .. code-block:: text

        word1 val1 val2 ... valM
        word2 val1 val2 ... valM
        ...
        wordN val1 val2 ... valM

    This function will iterate on each embedding in :obj:`embedding_file` and
    assign the pretrained vector to the associated word in :obj:`vocabulary_file`
    if found. Otherwise, the embedding is ignored.

    If :obj:`case_insensitive_embeddings` is ``True``, word embeddings are assumed
    to be trained on lowercase data. In that case, word alignments are case
    insensitive meaning the pretrained word embedding for "the" will be assigned
    to "the", "The", "THE", or any other case variants included in
    :obj:`vocabulary_file`.

    Args:
      embedding_file: Path the embedding file. Entries will be matched against
        :obj:`vocabulary_file`.
      vocabulary_file: The vocabulary file containing one word per line.
      num_oov_buckets: The number of additional unknown tokens.
      with_header: ``True`` if the embedding file starts with a header line like
        in GloVe embedding files.
      case_insensitive_embeddings: ``True`` if embeddings are trained on lowercase
        data.

    Returns:
      A Numpy array of shape ``[vocabulary_size + num_oov_buckets, embedding_size]``.
    """
    # Map words to ids from the vocabulary.
    word_to_id = collections.defaultdict(list)
    with tf.io.gfile.GFile(vocabulary_file) as vocabulary:
        count = 0
        for word in vocabulary:
            word = word.strip()
            if case_insensitive_embeddings:
                word = word.lower()
            word_to_id[word].append(count)
            count += 1

    # Fill pretrained embedding matrix.
    with tf.io.gfile.GFile(embedding_file) as embedding:
        pretrained = None

        if with_header:
            next(embedding)

        for line in embedding:
            fields = line.strip().split()
            word = fields[0]

            if pretrained is None:
                pretrained = np.random.normal(
                    size=(count + num_oov_buckets, len(fields) - 1)
                )

            # Lookup word in the vocabulary.
            if word in word_to_id:
                ids = word_to_id[word]
                for index in ids:
                    pretrained[index] = np.asarray(fields[1:])

    return pretrained


def add_sequence_controls(
    ids,
    length=None,
    start_id=None,
    end_id=None,
    pad_id=constants.PADDING_ID,
):
    """Adds sequence control tokens.

    Args:
      ids: Sequence of token ids as a 1D ``tf.Tensor``, a 2D ``tf.Tensor``, or a
        2D ``tf.RaggedTensor``.
      length: Length of each sequence in the batch as a 1D ``tf.Tensor``.
        Required if :obj:`ids` is a 2D ``tf.Tensor``.
      start_id: Token id to prepend to the sequences (set ``None`` to disable).
      end_id: Token id to append to the sequences (set ``None`` to disable).
      pad_id: Padding token ID.

    Returns:
      The updated token ids, and optionally the updated length if it was initially
      passed as argument.

    Raises:
      ValueError: if :obj:`ids` is a dense 2D ``tf.Tensor``, but :obj:`length` is not set.
      ValueError: if the rank of :obj:`ids` is not 1 or 2.
    """
    input_is_ragged = isinstance(ids, tf.RaggedTensor)
    if start_id is None and end_id is None:
        return ids if input_is_ragged or length is None else (ids, length)

    batch_size = None
    rank = ids.shape.rank
    if rank == 2:
        if not input_is_ragged:
            if length is None:
                raise ValueError(
                    "length argument is required when passing a dense batch of sequences"
                )
            ids = tf.RaggedTensor.from_tensor(ids, lengths=length)
        batch_size = ids.nrows()
    elif rank != 1:
        raise ValueError("Unsupported rank %d (expected 1 or 2)" % rank)

    def _make_column(value):
        value = tf.constant(value, dtype=ids.dtype)
        if batch_size is not None:
            value = tf.fill([batch_size], value)
        return tf.expand_dims(value, -1)

    concat_inputs = []
    if start_id is not None:
        concat_inputs.append(_make_column(start_id))
    concat_inputs.append(ids)
    if end_id is not None:
        concat_inputs.append(_make_column(end_id))

    ids = tf.concat(concat_inputs, axis=-1)

    if not input_is_ragged:
        if batch_size is not None:
            ids = ids.to_tensor(default_value=pad_id)
        if length is not None:
            length += len(concat_inputs) - 1
            return ids, length

    return ids


class TextInputter(Inputter):
    """An abstract inputter that processes text."""

    def __init__(self, num_oov_buckets=1, **kwargs):
        super().__init__(**kwargs)
        self.num_oov_buckets = num_oov_buckets
        self.noiser = None
        self.in_place_noise = False
        self.noise_probability = None
        self.vocabulary_file = None
        self.vocabulary_size = None
        self.tokens_to_ids = None
        self.ids_to_tokens = None
        self.tokenizer = None

    def initialize(self, data_config):
        self.vocabulary_file = data_config["vocabulary"]
        (
            self.vocabulary_size,
            self.tokens_to_ids,
            self.ids_to_tokens,
        ) = vocab.create_lookup_tables(
            self.vocabulary_file,
            self.num_oov_buckets,
            as_asset=data_config.get("export_vocabulary_assets", True),
        )
        tokenizer_config = data_config.get("tokenization")
        self.tokenizer = tokenizers.make_tokenizer(tokenizer_config)

    def set_noise(self, noiser, in_place=True, probability=None):
        """Enables noise to be applied to the input features.

        Args:
          noiser: A :class:`opennmt.data.WordNoiser` instance.
          in_place: If ``False``, the noisy version of the input will be stored as
            a separate feature prefixed with ``noisy_``.
          probability: When :obj:`in_place` is enabled, the probability to apply the
            noise.

        Raises:
          ValueError: if :obj:`in_place` is enabled but a :obj:`probability` is not
            set.
        """
        if in_place and probability is None:
            raise ValueError("In-place noise requires a probability")
        self.noiser = noiser
        self.in_place_noise = in_place
        self.noise_probability = probability

    def export_assets(self, asset_dir):
        self._assert_is_initialized()
        return self.tokenizer.export_assets(asset_dir, asset_prefix=self.asset_prefix)

    def make_dataset(self, data_file, training=None):
        return dataset_util.make_datasets(tf.data.TextLineDataset, data_file)

    def get_dataset_size(self, data_file):
        if isinstance(data_file, list):
            return list(map(misc.count_lines, data_file))
        return misc.count_lines(data_file)

    def has_prepare_step(self):
        # For performance reasons, we apply external tokenizers on a batch of
        # dataset elements during the preparation step.
        return not self.tokenizer.in_graph and not isinstance(
            self.tokenizer, tokenizers.SpaceTokenizer
        )

    def prepare_elements(self, elements, training=None):
        return {"tokens": self.tokenizer.tokenize(elements, training=training)}

    def make_features(self, element=None, features=None, training=None):
        """Tokenizes raw text."""
        self._assert_is_initialized()
        if features is None:
            features = {}
        if "tokens" in features:
            return features

        element = features.pop("text", element)
        if isinstance(element, dict):
            tokens = element["tokens"]
        else:
            element = tf.convert_to_tensor(element, dtype=tf.string)
            tokens = self.tokenizer.tokenize(element, training=training)

        if isinstance(tokens, tf.RaggedTensor):
            length = tokens.row_lengths()
            tokens = tokens.to_tensor(default_value=constants.PADDING_TOKEN)
        else:
            length = tf.shape(tokens)[0]
        if training and self.noiser is not None:
            noisy_tokens, noisy_length = self.noiser(
                tokens,
                sequence_length=length,
                probability=self.noise_probability if self.in_place_noise else None,
            )
            if self.in_place_noise:
                tokens = noisy_tokens
                length = noisy_length
            else:
                # Call make_features again to fill the remaining noisy features.
                noisy_features = dict(tokens=noisy_tokens, length=noisy_length)
                noisy_features = self.make_features(
                    features=noisy_features, training=training
                )
                for key, value in noisy_features.items():
                    features["noisy_%s" % key] = value
        features["length"] = length
        features["tokens"] = tokens
        return features

    def input_signature(self):
        self._assert_is_initialized()
        if self.tokenizer.in_graph:
            return {"text": tf.TensorSpec([None], tf.string)}
        else:
            return {
                "tokens": tf.TensorSpec([None, None], tf.string),
                "length": tf.TensorSpec([None], tf.int32),
            }

    def _assert_is_initialized(self):
        if self.tokenizer is None:
            raise RuntimeError(
                "The input layer is not initialized. You should initialize "
                "the model by calling model.initialize(data_config)."
            )


def _format_bool(b):
    return "yes" if b else "no"


class WordEmbedder(TextInputter):
    """Simple word embedder."""

    def __init__(self, embedding_size=None, dropout=0.0, **kwargs):
        """Initializes the parameters of the word embedder.

        Args:
          embedding_size: The size of the resulting embedding.
            If ``None``, an embedding file must be provided.
          dropout: The probability to drop units in the embedding.
          **kwargs: Additional layer keyword arguments.
        """
        super().__init__(**kwargs)
        self.embedding_size = embedding_size
        self.embedding_file = None
        self.dropout = dropout
        self.decoder_mode = False
        self.mark_start = None
        self.mark_end = None

    def set_decoder_mode(self, enable=True, mark_start=None, mark_end=None):
        """Make this inputter produce sequences for a decoder.

        In this mode, the returned "ids_out" feature is the decoder output sequence
        and "ids" is the decoder input sequence.

        Args:
          enable: Enable the decoder mode.
          mark_start: Mark the sequence start. If ``None``, keep the current value.
          mark_end: Mark the sequence end. If ``None``, keep the current value.
        """
        self.decoder_mode = enable
        if mark_start is not None:
            self.mark_start = mark_start
        if mark_end is not None:
            self.mark_end = mark_end

    def get_length(self, features, ignore_special_tokens=False):
        length = features["length"]
        if ignore_special_tokens:
            # Decoder mode shifts the sequences by one timesteps.
            num_special_tokens = -1 if self.decoder_mode else 0
            if self.mark_start:
                num_special_tokens += 1
            if self.mark_end:
                num_special_tokens += 1
            length -= num_special_tokens
        return length

    def get_oov_tokens(self, features):
        tokens, ids = features["tokens"], features["ids"]
        if self.mark_start:
            ids = ids[1:] if ids.shape.rank == 1 else ids[:, 1:]
        oov_tokens_pos = tf.where(ids >= (self.vocabulary_size - self.num_oov_buckets))
        return tf.gather_nd(tokens, oov_tokens_pos)

    def initialize(self, data_config):
        super().initialize(data_config)
        embedding = data_config.get("embedding")
        if embedding is None and self.embedding_size is None:
            raise ValueError("embedding_size must be set")
        if embedding is not None:
            self.embedding_file = embedding["path"]
            self.trainable = embedding.get("trainable", True)
            self.embedding_file_with_header = embedding.get("with_header", True)
            self.case_insensitive_embeddings = embedding.get("case_insensitive", True)
        sequence_controls = data_config.get("sequence_controls")
        if sequence_controls:
            self.mark_start = sequence_controls.get("start")
            self.mark_end = sequence_controls.get("end")

        tf.get_logger().info(
            "Initialized %s input layer:", self.asset_prefix.strip("_")
        )
        tf.get_logger().info(" - vocabulary size: %d", self.vocabulary_size)
        tf.get_logger().info(
            " - special tokens: BOS=%s, EOS=%s",
            _format_bool(self.mark_start),
            _format_bool(self.mark_end),
        )

    def make_features(self, element=None, features=None, training=None):
        """Converts words tokens to ids."""
        features = super().make_features(
            element=element, features=features, training=training
        )
        if "ids" not in features:
            features["ids"] = self.tokens_to_ids.lookup(features["tokens"])
            if self.mark_start or self.mark_end:
                features["ids"], features["length"] = add_sequence_controls(
                    features["ids"],
                    features["length"],
                    start_id=constants.START_OF_SENTENCE_ID
                    if self.mark_start
                    else None,
                    end_id=constants.END_OF_SENTENCE_ID if self.mark_end else None,
                )
        if self.decoder_mode:
            ids = features["ids"]
            length = features["length"]
            if ids.shape.rank == 2:
                mask = tf.sequence_mask(
                    length - 1, maxlen=tf.shape(ids)[1], dtype=ids.dtype
                )
                features["ids"] = (ids * mask)[:, :-1]
                features["ids_out"] = ids[:, 1:]
            else:
                features["ids"] = ids[:-1]
                features["ids_out"] = ids[1:]
            features["length"] -= 1
        return features

    def build(self, input_shape):
        if self.embedding_file:
            pretrained = load_pretrained_embeddings(
                self.embedding_file,
                self.vocabulary_file,
                num_oov_buckets=self.num_oov_buckets,
                with_header=self.embedding_file_with_header,
                case_insensitive_embeddings=self.case_insensitive_embeddings,
            )
            self.embedding_size = pretrained.shape[-1]
            initializer = tf.constant_initializer(value=pretrained.astype(self.dtype))
        else:
            initializer = None
        self.embedding = self.add_weight(
            "embedding",
            [self.vocabulary_size, self.embedding_size],
            initializer=initializer,
            trainable=self.trainable,
        )
        super().build(input_shape)

    def call(self, features, training=None):
        ids = features["ids"]

        if getattr(self, "_tflite_mode", False):
            # Workaround for TensorFlow issue #42410.
            ids = tf.expand_dims(ids, axis=-1)
            return tf.gather_nd(self.embedding, ids)

        outputs = tf.nn.embedding_lookup(self.embedding, ids)
        outputs = common.dropout(outputs, self.dropout, training=training)
        return outputs

    def visualize(self, model_root, log_dir):
        save_embeddings_metadata(
            log_dir,
            misc.get_variable_name(self.embedding, model_root),
            self.vocabulary_file,
            num_oov_buckets=self.num_oov_buckets,
        )

    def map_v1_weights(self, weights):
        return [(self.embedding, weights["w_embs"])]


class CharEmbedder(TextInputter):
    """Base class for character-aware inputters."""

    def __init__(self, embedding_size, dropout=0.0, **kwargs):
        """Initializes the parameters of the character embedder.

        Args:
          embedding_size: The size of the character embedding.
          dropout: The probability to drop units in the embedding.
          **kwargs: Additional layer keyword arguments.
        """
        super().__init__(**kwargs)
        self.embedding_size = embedding_size
        self.embedding = None
        self.dropout = dropout

    def make_features(self, element=None, features=None, training=None):
        """Converts words to characters."""
        if features is None:
            features = {}
        if "char_ids" in features:
            return features
        if "chars" in features:
            chars = features["chars"]
        else:
            features = super().make_features(
                element=element, features=features, training=training
            )
            chars = text.tokens_to_chars(features["tokens"])
            chars = chars.to_tensor(default_value=constants.PADDING_TOKEN)
        features["char_ids"] = self.tokens_to_ids.lookup(chars)
        return features

    def build(self, input_shape):
        self.embedding = self.add_weight(
            "char_embedding", [self.vocabulary_size, self.embedding_size]
        )
        super().build(input_shape)

    @abc.abstractmethod
    def call(self, features, training=None):
        raise NotImplementedError()

    def visualize(self, model_root, log_dir):
        save_embeddings_metadata(
            log_dir,
            misc.get_variable_name(self.embedding, model_root),
            self.vocabulary_file,
            num_oov_buckets=self.num_oov_buckets,
        )

    def _embed(self, inputs, training):
        mask = tf.math.not_equal(inputs, 0)
        outputs = tf.nn.embedding_lookup(self.embedding, inputs)
        outputs = common.dropout(outputs, self.dropout, training=training)
        return outputs, mask


class CharConvEmbedder(CharEmbedder):
    """An inputter that applies a convolution on characters embeddings."""

    def __init__(
        self,
        embedding_size,
        num_outputs,
        kernel_size=5,
        stride=3,
        dropout=0.0,
        **kwargs
    ):
        """Initializes the parameters of the character convolution embedder.

        Args:
          embedding_size: The size of the character embedding.
          num_outputs: The dimension of the convolution output space.
          kernel_size: Length of the convolution window.
          stride: Length of the convolution stride.
          dropout: The probability to drop units in the embedding.
          **kwargs: Additional layer keyword arguments.
        """
        super().__init__(embedding_size, dropout=dropout, **kwargs)
        self.output_size = num_outputs
        self.conv = tf.keras.layers.Conv1D(
            num_outputs, kernel_size, strides=stride, padding="same"
        )

    def call(self, features, training=None):
        inputs = features["char_ids"]
        flat_inputs = tf.reshape(inputs, [-1, tf.shape(inputs)[-1]])
        outputs, _ = self._embed(flat_inputs, training)
        outputs = self.conv(outputs)
        outputs = tf.reduce_max(outputs, axis=1)
        outputs = tf.reshape(outputs, [-1, tf.shape(inputs)[1], self.output_size])
        return outputs


class CharRNNEmbedder(CharEmbedder):
    """An inputter that runs a single RNN layer over character embeddings."""

    def __init__(
        self, embedding_size, num_units, dropout=0.2, cell_class=None, **kwargs
    ):
        """Initializes the parameters of the character RNN embedder.

        Args:
          embedding_size: The size of the character embedding.
          num_units: The number of units in the RNN layer.
          dropout: The probability to drop units in the embedding and the RNN
            outputs.
          cell_class: The inner cell class or a callable taking :obj:`num_units` as
            argument and returning a cell. Defaults to a LSTM cell.
          **kwargs: Additional layer keyword arguments.

        Raises:
          ValueError: if :obj:`encoding` is invalid.
        """
        super().__init__(embedding_size, dropout=dropout, **kwargs)
        if cell_class is None:
            cell_class = tf.keras.layers.LSTMCell
        self.rnn = tf.keras.layers.RNN(cell_class(num_units))
        self.num_units = num_units

    def call(self, features, training=None):
        inputs = features["char_ids"]
        flat_inputs = tf.reshape(inputs, [-1, tf.shape(inputs)[-1]])
        embeddings, mask = self._embed(flat_inputs, training)
        outputs = self.rnn(embeddings, mask=mask, training=training)
        outputs = tf.reshape(outputs, [-1, tf.shape(inputs)[1], self.num_units])
        return outputs



"""Sequence classifier."""

import tensorflow as tf

from opennmt import inputters
from opennmt.models.model import Model
from opennmt.utils.losses import cross_entropy_loss


class SequenceClassifier(Model):
    """A sequence classifier."""

    def __init__(self, inputter, encoder):
        """Initializes a sequence classifier.

        Args:
          inputter: A :class:`opennmt.inputters.Inputter` to process the
            input data.
          encoder: A :class:`opennmt.encoders.Encoder` to encode the input.

        Raises:
          ValueError: if :obj:`encoding` is invalid.
        """
        example_inputter = inputters.ExampleInputter(inputter, ClassInputter())
        super().__init__(example_inputter)
        self.encoder = encoder

    def build(self, input_shape):
        super().build(input_shape)
        self.output_layer = tf.keras.layers.Dense(self.labels_inputter.vocabulary_size)

    def call(self, features, labels=None, training=None, step=None):
        inputs = self.features_inputter(features, training=training)
        outputs, state, outputs_length = self.encoder(
            inputs,
            sequence_length=self.features_inputter.get_length(features),
            training=training,
        )

        if state is None:
            if outputs_length is not None:
                outputs = tf.RaggedTensor.from_tensor(outputs, lengths=outputs_length)
            encoding = tf.reduce_mean(outputs, axis=1)
        else:
            last_state = state[-1] if isinstance(state, (list, tuple)) else state
            encoding = (
                last_state if not isinstance(state, (list, tuple)) else last_state[0]
            )
        logits = self.output_layer(encoding)

        if not training:
            classes_prob = tf.nn.softmax(logits)
            classes_id = tf.argmax(classes_prob, axis=1)
            predictions = {
                "classes": self.labels_inputter.ids_to_tokens.lookup(classes_id),
                "classes_id": classes_id,
            }
        else:
            predictions = None

        return logits, predictions

    def compute_loss(self, outputs, labels, training=True):
        return cross_entropy_loss(
            outputs,
            labels["classes_id"],
            weight=labels.get("weight"),
            label_smoothing=self.params.get("label_smoothing", 0.0),
            training=training,
        )

    def get_metrics(self):
        return {"accuracy": tf.keras.metrics.Accuracy()}

    def update_metrics(self, metrics, predictions, labels):
        metrics["accuracy"].update_state(
            labels["classes_id"], predictions["classes_id"]
        )

    def format_prediction(self, prediction, params=None):
        return prediction["classes"].decode("utf-8")


class ClassInputter(inputters.TextInputter):
    """Reading class from a text file."""

    def __init__(self):
        super().__init__(num_oov_buckets=0)

    def make_features(self, element=None, features=None, training=None):
        if features is None:
            features = {}
        if "classes" not in features:
            features["classes"] = element
        features["classes_id"] = self.tokens_to_ids.lookup(features["classes"])
        return features

    def input_signature(self):
        return {"classes": tf.TensorSpec([None], tf.string)}



"""Define the Google's Transformer model."""

import tensorflow as tf

from opennmt import config as config_util
from opennmt import inputters
from opennmt.decoders.self_attention_decoder import SelfAttentionDecoder
from opennmt.encoders.encoder import ParallelEncoder
from opennmt.encoders.self_attention_encoder import SelfAttentionEncoder
from opennmt.layers.position import SinusoidalPositionEncoder
from opennmt.layers.transformer import MultiHeadAttentionReduction
from opennmt.models.sequence_to_sequence import (
    EmbeddingsSharingLevel,
    SequenceToSequence,
)


class Transformer(SequenceToSequence):
    """Attention-based sequence-to-sequence model as described in
    https://arxiv.org/abs/1706.03762.
    """

    def __init__(
        self,
        source_inputter=None,
        target_inputter=None,
        num_layers=6,
        num_units=512,
        num_heads=8,
        ffn_inner_dim=2048,
        dropout=0.1,
        attention_dropout=0.1,
        ffn_dropout=0.1,
        ffn_activation=tf.nn.relu,
        position_encoder_class=SinusoidalPositionEncoder,
        share_embeddings=EmbeddingsSharingLevel.NONE,
        share_encoders=False,
        maximum_relative_position=None,
        attention_reduction=MultiHeadAttentionReduction.FIRST_HEAD_LAST_LAYER,
        pre_norm=True,
    ):
        """Initializes a Transformer model.

        Args:
          source_inputter: A :class:`opennmt.inputters.Inputter` to process
            the source data. If this inputter returns parallel inputs, a multi
            source Transformer architecture will be constructed. Defaults to a
            :class:`opennmt.inputters.WordEmbedder` with :obj:`num_units` as
            embedding size.
          target_inputter: A :class:`opennmt.inputters.Inputter` to process
            the target data. Currently, only the
            :class:`opennmt.inputters.WordEmbedder` is supported. Defaults to a
            :class:`opennmt.inputters.WordEmbedder` with :obj:`num_units` as
            embedding size.
          num_layers: The number of layers or a 2-tuple with the number of encoder
            layers and decoder layers.
          num_units: The number of hidden units.
          num_heads: The number of heads in each self-attention layers.
          ffn_inner_dim: The inner dimension of the feed forward layers.
          dropout: The probability to drop units in each layer output.
          attention_dropout: The probability to drop units from the attention.
          ffn_dropout: The probability to drop units from the ReLU activation in
            the feed forward layer.
          ffn_activation: The activation function to apply between the two linear
            transformations of the feed forward layer.
          position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
            class to use for position encoding (or a callable that returns an
            instance).
          share_embeddings: Level of embeddings sharing, see
            :class:`opennmt.models.EmbeddingsSharingLevel` for possible values.
          share_encoders: In case of multi source architecture, whether to share the
            separate encoders parameters or not.
          maximum_relative_position: Maximum relative position representation
            (from https://arxiv.org/abs/1803.02155).
          attention_reduction: A :class:`opennmt.layers.MultiHeadAttentionReduction`
            value to specify how to reduce target-source multi-head attention
            matrices.
          pre_norm: If ``True``, layer normalization is applied before each
            sub-layer. Otherwise it is applied after. The original paper uses
            ``pre_norm=False``, but the authors later suggested that ``pre_norm=True``
            "seems better for harder-to-learn models, so it should probably be the
            default."
        """
        if source_inputter is None:
            source_inputter = inputters.WordEmbedder(embedding_size=num_units)
        if target_inputter is None:
            target_inputter = inputters.WordEmbedder(embedding_size=num_units)

        if isinstance(num_layers, (list, tuple)):
            num_encoder_layers, num_decoder_layers = num_layers
        else:
            num_encoder_layers, num_decoder_layers = num_layers, num_layers
        encoders = [
            SelfAttentionEncoder(
                num_encoder_layers,
                num_units=num_units,
                num_heads=num_heads,
                ffn_inner_dim=ffn_inner_dim,
                dropout=dropout,
                attention_dropout=attention_dropout,
                ffn_dropout=ffn_dropout,
                ffn_activation=ffn_activation,
                position_encoder_class=position_encoder_class,
                maximum_relative_position=maximum_relative_position,
                pre_norm=pre_norm,
            )
            for _ in range(source_inputter.num_outputs)
        ]
        if len(encoders) > 1:
            encoder = ParallelEncoder(
                encoders if not share_encoders else encoders[0],
                outputs_reducer=None,
                states_reducer=None,
            )
        else:
            encoder = encoders[0]
        decoder = SelfAttentionDecoder(
            num_decoder_layers,
            num_units=num_units,
            num_heads=num_heads,
            ffn_inner_dim=ffn_inner_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            ffn_dropout=ffn_dropout,
            ffn_activation=ffn_activation,
            position_encoder_class=position_encoder_class,
            num_sources=source_inputter.num_outputs,
            maximum_relative_position=maximum_relative_position,
            attention_reduction=attention_reduction,
            pre_norm=pre_norm,
        )

        self._pre_norm = pre_norm
        self._num_units = num_units
        self._num_encoder_layers = num_encoder_layers
        self._num_decoder_layers = num_decoder_layers
        self._num_heads = num_heads
        self._with_relative_position = maximum_relative_position is not None
        self._alignment_layer = -1
        self._alignment_heads = 1
        if attention_reduction == MultiHeadAttentionReduction.AVERAGE_LAST_LAYER:
            self._alignment_heads = 0
        self._is_ct2_compatible = (
            isinstance(encoder, SelfAttentionEncoder)
            and ffn_activation is tf.nn.relu
            and (
                (self._with_relative_position and position_encoder_class is None)
                or (
                    not self._with_relative_position
                    and position_encoder_class == SinusoidalPositionEncoder
                )
            )
        )
        super().__init__(
            source_inputter,
            target_inputter,
            encoder,
            decoder,
            share_embeddings=share_embeddings,
        )

    @property
    def ctranslate2_spec(self):
        if not self._is_ct2_compatible:
            return None
        import ctranslate2

        model_spec = ctranslate2.specs.TransformerSpec(
            (self._num_encoder_layers, self._num_decoder_layers),
            self._num_heads,
            with_relative_position=self._with_relative_position,
            pre_norm=self._pre_norm,
            alignment_layer=self._alignment_layer,
            alignment_heads=self._alignment_heads,
        )
        model_spec.with_source_bos = bool(self.features_inputter.mark_start)
        model_spec.with_source_eos = bool(self.features_inputter.mark_end)
        return model_spec

    def auto_config(self, num_replicas=1):
        config = super().auto_config(num_replicas=num_replicas)
        return config_util.merge_config(
            config,
            {
                "params": {
                    "average_loss_in_time": True,
                    "label_smoothing": 0.1,
                    "optimizer": "LazyAdam",
                    "optimizer_params": {"beta_1": 0.9, "beta_2": 0.998},
                    "learning_rate": 2.0,
                    "decay_type": "NoamDecay",
                    "decay_params": {
                        "model_dim": self._num_units,
                        "warmup_steps": 8000,
                    },
                },
                "train": {
                    "effective_batch_size": 25000,
                    "batch_size": 3072,
                    "batch_type": "tokens",
                    "maximum_features_length": (
                        100
                        if self.features_inputter.num_outputs == 1
                        else [100] * self.features_inputter.num_outputs
                    ),
                    "maximum_labels_length": 100,
                    "keep_checkpoint_max": 8,
                    "average_last_checkpoints": 8,
                },
            },
        )

    def map_v1_weights(self, weights):
        weights["seq2seq"] = weights.pop("transformer")
        return super().map_v1_weights(weights)



"""Define base tokenizers."""

import abc
import json
import sys

import tensorflow as tf
import yaml

from opennmt.utils import misc


class Tokenizer(abc.ABC):
    """Base class for tokenizers."""

    @property
    def in_graph(self):
        """Returns ``True`` if this tokenizer can be run in graph (i.e. uses TensorFlow ops)."""
        return False

    def export_assets(self, asset_dir, asset_prefix=""):
        """Exports assets for this tokenizer.

        Args:
          asset_dir: The directory where assets can be written.
          asset_prefix: The prefix to attach to assets filename.

        Returns:
          A dictionary containing additional assets used by the tokenizer.
        """
        return {}

    def tokenize_stream(
        self,
        input_stream=sys.stdin,
        output_stream=sys.stdout,
        delimiter=" ",
        training=True,
    ):
        """Tokenizes a stream of sentences.

        Args:
          input_stream: The input stream.
          output_stream: The output stream.
          delimiter: The token delimiter to use for text serialization.
          training: Set to ``False`` to tokenize for inference.
        """
        for line in input_stream:
            line = line.strip()
            tokens = self.tokenize(line, training=training)
            merged_tokens = delimiter.join(tokens)
            misc.print_as_bytes(merged_tokens, stream=output_stream)

    def detokenize_stream(
        self,
        input_stream=sys.stdin,
        output_stream=sys.stdout,
        delimiter=" ",
    ):
        """Detokenizes a stream of sentences.

        Args:
          input_stream: The input stream.
          output_stream: The output stream.
          delimiter: The token delimiter used for text serialization.
        """
        for line in input_stream:
            tokens = line.strip().split(delimiter)
            string = self.detokenize(tokens)
            misc.print_as_bytes(string, stream=output_stream)

    def tokenize(self, text, training=True):
        """Tokenizes text.

        Args:
          text: A string or batch of strings to tokenize as a ``tf.Tensor`` or
            Python values.
          training: Set to ``False`` to tokenize for inference.

        Returns:
          - If :obj:`text` is a Python string, a list of Python strings.
          - If :obj:`text` is a list of Python strings, a list of list of Python
            strings.
          - If :obj:`text` is a 0-D ``tf.Tensor``, a 1-D ``tf.Tensor``.
          - If :obj:`text` is a 1-D ``tf.Tensor``, a 2-D ``tf.RaggedTensor``.

        Raises:
          ValueError: if the rank of :obj:`text` is greater than 1.
        """
        with tf.device("cpu:0"):
            return self._tokenize(text, training)

    def _tokenize(self, text, training):
        if tf.is_tensor(text):
            return self._tokenize_tensor(text, training)
        elif isinstance(text, list):
            return list(map(lambda t: self._tokenize(t, training), text))
        else:
            text = tf.compat.as_text(text)
            return self._tokenize_string(text, training)

    def detokenize(self, tokens, sequence_length=None):
        """Detokenizes tokens.

        The Tensor version supports batches of tokens.

        Args:
          tokens: Tokens or batch of tokens as a ``tf.Tensor``, ``tf.RaggedTensor``,
            or Python values.
          sequence_length: The length of each sequence. Required if :obj:`tokens`
            is a dense 2-D ``tf.Tensor``.

        Returns:
          - If :obj:`tokens` is a list of list of Python strings, a list of Python strings.
          - If :obj:`tokens` is a list of Python strings, a Python string.
          - If :obj:`tokens` is a N-D ``tf.Tensor`` (or ``tf.RaggedTensor``), a
            (N-1)-D ``tf.Tensor``.

        Raises:
          ValueError: if the rank of :obj:`tokens` is greater than 2.
          ValueError: if :obj:`tokens` is a 2-D dense ``tf.Tensor`` and
            :obj:`sequence_length` is not set.
        """
        with tf.device("cpu:0"):
            return self._detokenize(tokens, sequence_length)

    def _detokenize(self, tokens, sequence_length):
        if isinstance(tokens, tf.RaggedTensor):
            return self._detokenize_tensor(tokens)
        elif tf.is_tensor(tokens):
            rank = len(tokens.shape)
            if rank == 1:
                return self._detokenize_tensor(tokens)
            elif rank == 2:
                if sequence_length is None:
                    raise ValueError(
                        "sequence_length is required for Tensor detokenization"
                    )
                tokens = tf.RaggedTensor.from_tensor(tokens, lengths=sequence_length)
                return self._detokenize_tensor(tokens)
            else:
                raise ValueError("Unsupported tensor rank %d for detokenization" % rank)
        elif isinstance(tokens, list) and tokens and isinstance(tokens[0], list):
            return list(map(self.detokenize, tokens))
        else:
            tokens = [tf.compat.as_text(token) for token in tokens]
            return self._detokenize_string(tokens)

    def _tokenize_tensor(self, text, training):
        """Tokenizes a tensor.

        When not overriden, this default implementation calls the string-based
        tokenization.

        Args:
          text: A 0-D or 1-D string ``tf.Tensor``.
          training: Set to ``False`` to tokenize for inference.

        Returns:
          A 1-D string ``tf.Tensor``, or a 2-D string ``tf.RaggedTensor`` if the
          input was a batch of text.
        """

        def _python_wrapper(string_t):
            string = tf.compat.as_text(string_t.numpy())
            tokens = self._tokenize_string(string, training)
            return tf.constant(tokens, dtype=tf.string)

        def _python_wrapper_batch(batch_text):
            batch_text = list(map(tf.compat.as_text, batch_text.numpy()))
            batch_tokens = self._tokenize_string_batch(batch_text, training)
            flat_tokens = tf.constant(tf.nest.flatten(batch_tokens), dtype=tf.string)
            lengths = tf.constant(list(map(len, batch_tokens)), dtype=tf.int32)
            return flat_tokens, lengths

        rank = text.shape.rank
        if rank == 0:
            tokens = tf.py_function(_python_wrapper, [text], tf.string)
            tokens.set_shape([None])
            return tokens
        elif rank == 1:
            flat_tokens, lengths = tf.py_function(
                _python_wrapper_batch, [text], (tf.string, tf.int32)
            )
            flat_tokens.set_shape([None])
            lengths.set_shape([None])
            return tf.RaggedTensor.from_row_lengths(flat_tokens, lengths)
        else:
            raise ValueError("Unsupported tensor rank %d for tokenization" % rank)

    def _detokenize_tensor(self, tokens):
        """Detokenizes tokens.

        When not overriden, this default implementation calls the string-based
        detokenization.

        Args:
          tokens: A 1-D string ``tf.Tensor``, or a 2-D string ``tf.RaggedTensor``.

        Returns:
          A 0-D string ``tf.Tensor``, or a 1-D string ``tf.Tensor`` if the input
          was a batch of tokens.
        """

        def _python_wrapper(tokens_t):
            tokens = [tf.compat.as_text(s) for s in tokens_t.numpy()]
            string = self._detokenize_string(tokens)
            return tf.constant(string)

        rank = tokens.shape.rank
        if rank == 1:
            text = tf.py_function(_python_wrapper, [tokens], tf.string)
            text.set_shape([])
            return text
        elif rank == 2:
            return tf.map_fn(self._detokenize_tensor, tokens, dtype=tf.string)
        else:
            raise ValueError("Unsupported tensor rank %d for detokenization" % rank)

    @abc.abstractmethod
    def _tokenize_string(self, text, training):
        """Tokenizes a Python unicode string.

        This method should be thread-safe.

        Args:
          text: A Python unicode string.
          training: Set to ``False`` to tokenize for inference.

        Returns:
          A list of Python unicode strings.
        """
        raise NotImplementedError()

    def _tokenize_string_batch(self, batch_text, training):
        """Tokenizes a batch of Python unicode strings.

        Args:
          batch_text: A list of Python unicode strings.
          training: Set to ``False`` to tokenize for inference.

        Returns:
          A list of lists of Python unicode strings.
        """
        return [self._tokenize_string(text, training) for text in batch_text]

    @abc.abstractmethod
    def _detokenize_string(self, tokens):
        """Detokenizes tokens.

        Args:
          tokens: A list of Python unicode strings.

        Returns:
          A unicode Python string.
        """
        raise NotImplementedError()


_TOKENIZERS_REGISTRY = misc.ClassRegistry(base_class=Tokenizer)

register_tokenizer = _TOKENIZERS_REGISTRY.register


def make_tokenizer(config=None):
    """Creates a tokenizer instance from the configuration.

    Args:
      config: Tokenization configuration as a Python dictionary, or a path to a
        YAML configuration file, or a JSON string.

    Returns:
      A :class:`opennmt.tokenizers.Tokenizer` instance.

    Raises:
      ValueError: if :obj:`config` is invalid.
    """
    if config:
        if isinstance(config, str):
            if tf.io.gfile.exists(config):
                with tf.io.gfile.GFile(config) as config_file:
                    config = yaml.safe_load(config_file)
            else:
                try:
                    config = json.loads(config)
                except json.JSONDecodeError:
                    pass
        if isinstance(config, dict):
            tokenizer_type = config.get("type")
            if tokenizer_type is None:
                tokenizer_type = "OpenNMTTokenizer"
                tokenizer_params = config
            else:
                tokenizer_params = config.get("params", {})
            tokenizer_class = _TOKENIZERS_REGISTRY.get(tokenizer_type)
            if tokenizer_class is None:
                raise ValueError(
                    "%s is not in list of accepted tokenizers: %s"
                    % (
                        tokenizer_type,
                        ", ".join(sorted(_TOKENIZERS_REGISTRY.class_names)),
                    )
                )
            tokenizer = tokenizer_class(**tokenizer_params)
        else:
            raise ValueError("Invalid tokenization configuration: %s" % str(config))
    else:
        # If the tokenization was not configured, we assume that an external tokenization
        # was used and we don't include the tokenizer in the exported graph.
        tokenizer = SpaceTokenizer(in_graph=False)
    return tokenizer


def _process_stream_as_dataset(
    input_stream,
    output_stream,
    map_func,
    batch_size=512,
    num_parallel_calls=4,
):
    dataset = tf.data.Dataset.from_generator(
        lambda: input_stream,
        output_types=tf.string,
        output_shapes=tf.TensorShape([]),
    )
    dataset = dataset.batch(batch_size)
    dataset = dataset.map(map_func, num_parallel_calls=num_parallel_calls)

    expected_spec = tf.TensorSpec(shape=[None], dtype=tf.string)
    if dataset.element_spec != expected_spec:
        raise TypeError(
            "Expected map_func to produce elements with spec %s, but got spec %s instead"
            % (expected_spec, dataset.element_spec)
        )

    for lines in dataset.as_numpy_iterator():
        for line in lines:
            misc.print_as_bytes(line, stream=output_stream)


class TensorFlowTokenizer(tf.Module, Tokenizer):
    """Base class for tokenizers using only TensorFlow ops."""

    @property
    def in_graph(self):
        return True

    def tokenize_stream(
        self,
        input_stream=sys.stdin,
        output_stream=sys.stdout,
        delimiter=" ",
        training=True,
    ):
        def _map_func(line):
            line = tf.strings.strip(line)
            tokens = self._tokenize_tensor(line, training)
            return tf.strings.reduce_join(
                tokens, axis=tokens.shape.rank - 1, separator=delimiter
            )

        _process_stream_as_dataset(input_stream, output_stream, _map_func)

    def detokenize_stream(
        self,
        input_stream=sys.stdin,
        output_stream=sys.stdout,
        delimiter=" ",
    ):
        def _map_func(line):
            line = tf.strings.strip(line)
            tokens = tf.strings.split(line, sep=delimiter)
            return self._detokenize_tensor(tokens)

        _process_stream_as_dataset(input_stream, output_stream, _map_func)

    @abc.abstractmethod
    def _tokenize_tensor(self, text, training):
        raise NotImplementedError()

    @abc.abstractmethod
    def _detokenize_tensor(self, tokens):
        raise NotImplementedError()

    def _tokenize_string(self, text, training):
        tokens = self._tokenize_tensor(tf.constant(text, dtype=tf.string), training)
        return [token.decode("utf-8") for token in tokens.numpy()]

    def _detokenize_string(self, tokens):
        text = self._detokenize_tensor(tf.constant(tokens, dtype=tf.string))
        return text.numpy().decode("utf-8")


@register_tokenizer
class SpaceTokenizer(Tokenizer):
    """A tokenizer that splits on spaces."""

    def __init__(self, in_graph=True):
        """Initializes the tokenizer.

        Args:
          in_graph: If ``True``, this tokenizer should be integrated in the exported graph.
        """
        self._in_graph = in_graph

    @property
    def in_graph(self):
        return self._in_graph

    def _tokenize_tensor(self, text, _):
        return tf.strings.split(text)

    def _detokenize_tensor(self, tokens):
        return tf.strings.reduce_join(tokens, axis=tokens.shape.rank - 1, separator=" ")

    def _tokenize_string(self, text, _):
        return text.split()

    def _detokenize_string(self, tokens):
        return " ".join(tokens)


@register_tokenizer
class CharacterTokenizer(Tokenizer):
    """A tokenizer that splits unicode characters."""

    @property
    def in_graph(self):
        return True

    def _tokenize_tensor(self, text, _):
        text = tf.strings.regex_replace(text, " ", "▁")
        return tf.strings.unicode_split(text, "UTF-8")

    def _detokenize_tensor(self, tokens):
        text = tf.strings.reduce_join(tokens, axis=tokens.shape.rank - 1)
        return tf.strings.regex_replace(text, "▁", " ")

    def _tokenize_string(self, text, _):
        return list(text.replace(" ", "▁"))

    def _detokenize_string(self, tokens):
        return "".join(tokens).replace("▁", " ")



"""Base class for models."""

import abc
import contextlib

import tensorflow as tf

from opennmt import optimizers, schedules
from opennmt.utils import exporters, losses, misc


class Model(tf.keras.layers.Layer):
    """Base class for models."""

    def __init__(self, examples_inputter):
        super().__init__()
        self.examples_inputter = examples_inputter
        self.params = {}

    @property
    def unsupervised(self):
        """Unsupervised model."""
        return self.labels_inputter is None

    @property
    def features_inputter(self):
        """The inputter producing features."""
        return getattr(
            self.examples_inputter, "features_inputter", self.examples_inputter
        )

    @property
    def labels_inputter(self):
        """The inputter producing labels."""
        return getattr(self.examples_inputter, "labels_inputter", None)

    @property
    def ctranslate2_spec(self):
        """The equivalent CTranslate2 model specification."""
        return None

    def __repr__(self):
        """Returns a description of the model and its submodules."""
        return misc.describe_layer(self, name="model")

    def auto_config(self, num_replicas=1):
        """Returns automatic configuration values specific to this model.

        Args:
          num_replicas: The number of synchronous model replicas used for the
            training.

        Returns:
          A partial training configuration.
        """
        _ = num_replicas
        return {}

    def initialize(self, data_config, params=None):
        """Initializes the model from the data configuration.

        Args:
          data_config: A dictionary containing the data configuration set
            by the user (e.g. vocabularies, tokenization, pretrained embeddings,
            etc.).
          params: A dictionary of hyperparameters.
        """
        if params is None:
            params = {}
        self.params.update(params)
        dropout = self.params.get("dropout")
        if dropout is not None:
            misc.set_dropout(self, dropout)
        self.examples_inputter.initialize(data_config)

    def build(self, input_shape):
        freeze_layers = self.params.get("freeze_layers")
        if freeze_layers:
            if not isinstance(freeze_layers, list):
                freeze_layers = [freeze_layers]
            for layer_path in freeze_layers:
                layer = misc.index_structure(self, layer_path)
                layer.trainable = False
                misc.set_dropout(layer, 0)  # Disable dropout in frozen layers.
        self.examples_inputter.build(input_shape)
        self.built = True

    def __call__(self, features, labels=None, training=None, step=None):
        """Runs the model.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.
          training: If ``True``, run in training mode.
          step: The current training step.

        Returns:
          A tuple containing,

          - The model outputs (usually unscaled probabilities).
          - The model predictions.
        """
        outputs, predictions = super().__call__(
            features,
            labels=labels,
            training=training,
            step=step,
        )

        # Include the example index vector in the outputs.
        index = features.get("index") if isinstance(features, dict) else None
        if index is not None:
            if isinstance(outputs, dict):
                outputs["index"] = index
            if isinstance(predictions, dict):
                predictions["index"] = index

        return outputs, predictions

    @abc.abstractmethod
    def call(self, features, labels=None, training=None, step=None):
        """Runs the model.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.
          training: If ``True``, run in training mode.
          step: The current training step.

        Returns:
          A tuple containing,

          - The model outputs (usually unscaled probabilities).
          - The model predictions.
        """
        raise NotImplementedError()

    def infer(self, features):
        """Runs inference on :obj:`features`.

        This is a small convenience wrapper around
        :meth:`opennmt.models.Model.call`.

        Args:
          features: A nested structure of features ``tf.Tensor``.

        Returns:
          The model predictions.
        """
        _, predictions = self(features)
        return predictions

    def evaluate(self, features, labels):
        """Evaluates :obj:`features` predictions against `labels`.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.

        Returns:
          A tuple with the loss and the model predictions.
        """
        outputs, predictions = self(features, labels=labels)
        loss = self.compute_loss(outputs, labels, training=False)
        return loss, predictions

    def score(self, features, labels):
        """Scores labels.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.

        Returns:
          The score results.
        """
        raise NotImplementedError("This model does not define a score function")

    def train(self, features, labels, optimizer, loss_scale=None):
        """Computes and applies the gradients for a batch of examples.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.
          optimizer: The optimizer instance
            (``tf.keras.mixed_precision.LossScaleOptimizer`` is supported).
          loss_scale: An optional loss scaling factor.

        Returns:
          The loss.
        """
        loss, gradients = self.compute_gradients(
            features,
            labels,
            optimizer,
            loss_scale=loss_scale,
        )
        optimizer.apply_gradients(list(zip(gradients, self.trainable_weights)))
        return loss

    def compute_gradients(self, features, labels, optimizer, loss_scale=None):
        """Computes the gradients for a batch of examples.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.
          optimizer: The optimizer instance
            (``tf.keras.mixed_precision.LossScaleOptimizer`` is supported).
          loss_scale: An optional loss scaling factor.

        Returns:
          A tuple containing,

          - The loss.
          - The gradients.
        """

        def _compute_loss():
            train_loss, report_loss = self.compute_training_loss(
                features,
                labels,
                step=optimizer.iterations,
            )
            if loss_scale is not None:
                train_loss /= loss_scale
                report_loss /= loss_scale
            return train_loss, report_loss

        if tf.executing_eagerly():
            with tf.GradientTape() as tape:
                train_loss, report_loss = _compute_loss()
                if isinstance(optimizer, tf.keras.mixed_precision.LossScaleOptimizer):
                    train_loss = optimizer.get_scaled_loss(train_loss)
            gradients = tape.gradient(train_loss, self.trainable_weights)
            if isinstance(optimizer, tf.keras.mixed_precision.LossScaleOptimizer):
                gradients = optimizer.get_unscaled_gradients(gradients)

        else:
            train_loss, report_loss = _compute_loss()
            # LossScaleOptimizer.get_gradients takes care of loss scaling.
            gradients = optimizer.get_gradients(train_loss, self.trainable_weights)

        return report_loss, gradients

    def compute_training_loss(self, features, labels, step=None):
        """Computes the training loss for a batch of examples.

        Args:
          features: A nested structure of features ``tf.Tensor``.
          labels: A nested structure of labels ``tf.Tensor``.
          step: The current training step.

        Returns:
          A tuple containing,

          - The loss to optimize.
          - The loss to report.
        """
        outputs, _ = self(features, labels, training=True, step=step)
        loss = self.compute_loss(outputs, labels, training=True)
        if isinstance(loss, tuple):
            train_loss = loss[0] / loss[1]
            report_loss = loss[0] / loss[2] if len(loss) > 2 else train_loss
        else:
            train_loss, report_loss = loss, loss
        train_loss = self.regularize_loss(train_loss, variables=self.trainable_weights)
        return train_loss, report_loss

    @abc.abstractmethod
    def compute_loss(self, outputs, labels, training=True):
        """Computes the loss.

        Args:
          outputs: The model outputs (usually unscaled probabilities).
          labels: The dict of labels ``tf.Tensor``.
          training: If ``True``, compute the loss for training.

        Returns:
          The loss or a tuple ``(numerator, train_denominator, stats_denominator)``
          to use a different normalization for training compared to reporting (e.g.
          batch-normalized for training vs. token-normalized for reporting).
        """
        raise NotImplementedError()

    def regularize_loss(self, loss, variables=None):
        """Regularizes the loss.

        Args:
          loss: The loss.
          variables: List of variables.

        Returns:
          The regularized loss.
        """
        if variables is None:
            variables = self.trainable_variables
        regularization = self.params.get("regularization")
        if regularization is not None:
            loss += losses.regularization_penalty(
                regularization["type"], regularization["scale"], variables
            )
        return loss

    def get_metrics(self):
        """Returns the metrics for this model.

        Returns:
          A dictionary of ``tf.keras.metrics.Metric`` metrics.
        """
        return None

    def update_metrics(self, metrics, predictions, labels):
        """Computes additional metrics on the predictions.

        Args:
          metrics: A dictionary of metrics to update.
          predictions: The model predictions.
          labels: The dict of labels ``tf.Tensor``.
        """
        return

    def get_optimizer(self):
        """Returns the optimizer for this model.

        Returns:
          A ``tf.keras.optimizers.Optimizer`` instance or ``None`` if no optimizer
          is configured.
        """
        params = self.params
        optimizer_name = params.get("optimizer")
        if optimizer_name is None:
            return None
        schedule_type = params.get("decay_type")
        if schedule_type is None:
            learning_rate = tf.constant(params["learning_rate"], dtype=tf.float32)
        else:
            schedule_params = params.get("decay_params", {})
            learning_rate = schedules.make_learning_rate_schedule(
                params.get("learning_rate"),
                schedule_type,
                schedule_params=schedule_params,
                schedule_step_duration=params.get("decay_step_duration", 1),
                start_step=params.get("start_decay_steps", 0),
                minimum_learning_rate=params.get("minimum_learning_rate", 0),
            )
        optimizer_params = params.get("optimizer_params")
        if optimizer_params is None:
            optimizer_params = {}
        optimizer = optimizers.make_optimizer(
            optimizer_name, learning_rate, **optimizer_params
        )
        return optimizer

    def serve_function(self):
        """Returns a function for serving this model.

        Returns:
          A ``tf.function``.
        """
        # Set name attribute of the input TensorSpec.
        input_signature = {
            name: tf.TensorSpec.from_spec(spec, name=name)
            for name, spec in self.features_inputter.input_signature().items()
        }

        @tf.function(input_signature=(input_signature,))
        def _run(features):
            features = self.features_inputter.make_features(features=features.copy())
            _, predictions = self(features)
            return predictions

        return _run

    @property
    def tflite_mode(self):
        """Returns ``True`` if the model is being traced for TensorFlow Lite."""
        return getattr(self, "_tflite_mode", False)

    @contextlib.contextmanager
    def enable_tflite_mode(self):
        """Enable TensorFlow Lite mode for this model."""
        layers = [self] + list(self.submodules)
        for layer in layers:
            setattr(layer, "_tflite_mode", True)
        yield
        for layer in layers:
            delattr(layer, "_tflite_mode")

    def tflite_function(self):
        """Returns the inference function that should be used for TensorFlow Lite.

        Returns:
          A ``tf.function``.
        """
        with self.enable_tflite_mode():
            return self.serve_function()

    def export(self, export_dir, exporter=None):
        """Exports the model for serving.

        Args:
          export_dir: The output directory.
          exporter: A :class:`opennmt.utils.Exporter` instance. Defaults to
            :class:`opennmt.utils.SavedModelExporter`.
        """
        if exporter is None:
            exporter = exporters.SavedModelExporter()
        exporter.export(self, export_dir)

    def create_variables(self, optimizer=None):
        """Creates the model variables by running it once.

        Args:
          optimizer: If set, also create the optimizer variables.
        """
        # Create input features from the input signatures. We remove the leading
        # batch dimension as sometimes assumed by make_features methods and set
        # unspecified dimensions to 1.
        features = tf.nest.map_structure(
            lambda spec: tf.fill(
                [dim or 1 for dim in spec.shape.as_list()[1:]],
                tf.constant("a" if spec.dtype is tf.string else 1, dtype=spec.dtype),
            ),
            self.examples_inputter.input_signature(),
        )
        features = self.examples_inputter.make_features(features=features)

        # Add the batch dimension back before calling the model.
        features, labels = tf.nest.map_structure(
            lambda x: tf.expand_dims(x, 0), features
        )
        _ = self(features, labels=labels, training=True, step=0)

        if optimizer is not None:
            optimizer._create_all_weights(self.trainable_variables)

    def transfer_weights(
        self, new_model, new_optimizer=None, optimizer=None, ignore_weights=None
    ):
        """Transfers weights (and optionally optimizer slots) from this model to
        another.

        This default implementation assumes that :obj:`self` and :obj:`new_model`
        have exactly the same variables. Subclasses can override this method to
        transfer weights to another model type or architecture. For example,
        :class:`opennmt.models.SequenceToSequence` can transfer weights to a model
        with a different vocabulary.

        All model and optimizer variables are expected to be initialized.

        Args:
          new_model: The new model to transfer weights to.
          new_optimizer: The new optimizer.
          optimizer: The optimizer used for the current model.
          ignore_weights: Optional list of weights to not transfer.
        """
        if type(self) is not type(new_model):
            raise ValueError(
                "Transferring weights to another model type is not supported"
            )
        if ignore_weights is None:
            ignore_weights = set()
        ignore_weights_ref = set(weight.ref() for weight in ignore_weights)
        weights = self.weights
        new_weights = new_model.weights
        for weight, new_weight in zip(weights, new_weights):
            if new_weight.ref() not in ignore_weights_ref:
                new_weight.assign(weight)
                if new_optimizer is not None and optimizer is not None:
                    for slot_name in new_optimizer.get_slot_names():
                        if slot_name not in optimizer.get_slot_names():
                            continue
                        new_slot = new_optimizer.get_slot(new_weight, slot_name)
                        slot = optimizer.get_slot(weight, slot_name)
                        new_slot.assign(slot)

    def map_v1_weights(self, weights):
        """Maps current weights to V1 weights.

        Args:
          weights: A nested dictionary following the scope names used in V1. The
            leaves are tuples with the variable value and optionally the optimizer
            slots.

        Returns:
          A list of tuples associating variables and their V1 equivalent.
        """
        raise NotImplementedError("This model can not restore V1 checkpoints")

    def export_assets(self, asset_dir):
        """Exports additional assets used by this model.

        Args:
          asset_dir: The directory where assets can be written.

        Returns:
          A dictionary of additional assets.
        """
        return self.examples_inputter.export_assets(asset_dir)

    def visualize(self, log_dir):
        """Setups model visualization (e.g. word embedding projections).

        Args:
          log_dir: The log directory.
        """
        self.features_inputter.visualize(self, log_dir)
        if not self.unsupervised:
            self.labels_inputter.visualize(self, log_dir)

    def format_prediction(self, prediction, params=None):
        """Formats the model prediction for file saving.

        Args:
          prediction: The model prediction (same structure as the second output of
            :meth:`opennmt.models.Model.call`).
          params: (optional) Dictionary of formatting parameters.

        Returns:
          A string or list of strings.
        """
        return str(prediction)

    def format_score(self, score, params=None, stream=None):
        """Formats the score result for file saving.

        Args:
          score: The score result (same structure as the output of
            :meth:`opennmt.models.Model.score`).
          params: (optional) Dictionary of formatting parameters.
        """
        return str(score)

    def print_prediction(self, prediction, params=None, stream=None):
        """Prints the model prediction.

        Args:
          prediction: The model prediction (same structure as the second output of
            :meth:`opennmt.models.Model.call`).
          params: (optional) Dictionary of formatting parameters.
          stream: (optional) The stream to print to.
        """
        _write_lines(self.format_prediction(prediction, params=params), stream)

    def print_score(self, score, params=None, stream=None):
        """Prints the score result.

        Args:
          score: The score result (same structure as the output of
            :meth:`opennmt.models.Model.score`).
          params: (optional) Dictionary of formatting parameters.
          stream: (optional) The stream to print to.
        """
        _write_lines(self.format_score(score, params=params), stream)


def _write_lines(lines, stream):
    if not isinstance(lines, list):
        lines = [lines]
    for line in lines:
        misc.print_as_bytes(line, stream=stream)


class SequenceGenerator(Model):
    """Base class for models generating sequences."""

    @property
    def decoder_inputter(self):
        """The inputter used on the decoder side."""
        return self.labels_inputter if not self.unsupervised else self.features_inputter

    def score(self, features, labels):
        outputs, _ = self(features, labels=labels)
        cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels["ids_out"], outputs["logits"]
        )
        weights = tf.sequence_mask(labels["length"], dtype=cross_entropy.dtype)
        masked_cross_entropy = cross_entropy * weights
        scores = tf.reduce_sum(masked_cross_entropy, axis=1)
        results = {
            "cross_entropy": cross_entropy,
            "score": scores,
            "tokens": labels["tokens"],
            "length": self.decoder_inputter.get_length(
                labels, ignore_special_tokens=True
            ),
        }
        for key_to_forward in ("attention", "index"):
            value = outputs.get(key_to_forward)
            if value is not None:
                results[key_to_forward] = value
        return results

    def format_score(self, score, params=None):
        if params is None:
            params = {}
        length = score["length"]
        tokens = score["tokens"][:length]
        sentence = " ".join(token.decode("utf-8") for token in tokens)
        token_level_scores = None
        attention = None
        if params.get("with_token_level"):
            token_level_scores = score["cross_entropy"][:length]
        if "attention" in score:
            attention = score["attention"][:length]
        alignment_type = params.get("with_alignments")
        return misc.format_translation_output(
            sentence,
            score=score["score"],
            token_level_scores=token_level_scores,
            attention=attention,
            alignment_type=alignment_type,
        )



"""Dataset creation and transformations."""

import numpy as np
import tensorflow as tf

from opennmt.utils import compat, misc


def make_datasets(dataset_cls, filenames):
    """Creates instances of :obj:`dataset_cls`.

    Args:
      dataset_cls: A class inheriting from ``tf.data.Dataset``.
      filenames: A list of filenames or a single filename.

    Returns:
      A list of ``tf.data.Dataset`` instances if multiple :obj:`filenames` are
      passed, otherwise a single ``tf.data.Dataset``.

    Raises:
      ValueError: if :obj:`filenames` is empty.
    """
    if not isinstance(filenames, list):
        filenames = [filenames]
    elif not filenames:
        raise ValueError("At least one data file is required")
    datasets = [
        dataset_cls(
            filename, compression_type="GZIP" if misc.is_gzip_file(filename) else None
        )
        for filename in filenames
    ]
    if len(datasets) == 1:
        return datasets[0]
    return datasets


def normalize_weights(datasets, weights=None, sizes=None):
    """Returns normalized dataset weights based on datasets sizes.

    Args:
      datasets: A list of ``tf.data.Dataset`` instances.
      weights: An optional list of dataset weights.
      sizes: The size of each dataset, if known.

    Returns:
      A normalized list of weights that can be used as sampling probabilities.

    Raises:
      ValueError: if the length of :obj:`weights` or :obj:`sizes` does not match
        the length of :obj:`datasets`.
    """
    if not datasets:
        return []
    if len(datasets) == 1:
        return [1.0]

    if weights is None:
        weights = [1 / len(datasets)] * len(datasets)
    elif len(weights) != len(datasets):
        raise ValueError(
            "Got %d datasets but %d weights" % (len(datasets), len(weights))
        )

    if sizes is None:
        sizes = [int(get_dataset_size(dataset)) for dataset in datasets]
    elif len(sizes) != len(datasets):
        raise ValueError("Got %d datasets but %d sizes" % (len(datasets), len(sizes)))

    # Weights should be normalized by the dataset size relative to the total size.
    total_size = sum(sizes)
    weights = [weight * (size / total_size) for weight, size in zip(weights, sizes)]

    # Convert weights to probabilities.
    total_weights = sum(weights)
    return [weight / total_weights for weight in weights]


def _get_output_shapes(dataset):
    """Returns the outputs shapes of the dataset.

    Args:
      dataset: A ``tf.data.Dataset``.

    Returns:
      A nested structure of ``tf.TensorShape``
    """
    return tf.nest.map_structure(lambda spec: spec.shape, dataset.element_spec)


def get_dataset_size(dataset, batch_size=5000):
    """Get the dataset size.

    Example:

      >>> dataset = tf.data.Dataset.range(5)
      >>> opennmt.data.get_dataset_size(dataset).numpy()
      5

    Args:
      dataset: A dataset.
      batch_size: The batch size to use to improve the scan performance, or
        ``None`` to scan the dataset as-is.

    Returns:
      The dataset size or ``None`` if the dataset is infinite.
    """
    if dataset.cardinality() == tf.data.INFINITE_CARDINALITY:
        return None
    if batch_size is not None:
        dataset = dataset.batch(batch_size)

    def _reduce_func(count, element):
        element = tf.nest.flatten(element)[0]
        batch_size = tf.shape(element)[0]
        return count + tf.cast(batch_size, count.dtype)

    return dataset.reduce(tf.constant(0, dtype=tf.int64), _reduce_func)


def filter_irregular_batches(multiple):
    """Transformation that filters out batches based on their size.

    Example:

      >>> dataset = tf.data.Dataset.range(10).batch(3)
      >>> dataset = dataset.apply(opennmt.data.filter_irregular_batches(3))
      >>> len(list(iter(dataset)))
      3

    Args:
      multiple: The divisor of the batch size.

    Returns:
      A ``tf.data.Dataset`` transformation.
    """
    if multiple == 1:
        return lambda dataset: dataset

    def _predicate(*x):
        flat = tf.nest.flatten(x)
        batch_size = tf.shape(flat[0])[0]
        return tf.equal(batch_size % multiple, 0)

    return lambda dataset: dataset.filter(_predicate)


def filter_examples_by_length(
    maximum_features_length=None,
    maximum_labels_length=None,
    features_length_fn=None,
    labels_length_fn=None,
):
    """Transformation that filters out examples with zero length or length that is
    greater than the configured maximum.

    Example:

      >>> dataset = dataset.apply(opennmt.data.filter_examples_by_length(...))

    Args:
      maximum_features_length: The maximum length or list of maximum lengths of
        the features sequence(s). ``None`` to not constrain the length.
      maximum_labels_length: The maximum length or list of maximum lengths of
        the labels sequence(s). ``None`` to not constrain the length.
      features_length_fn: A function mapping features to a sequence length.
      labels_length_fn: A function mapping labels to a sequence length.

    Returns:
      A ``tf.data.Dataset`` transformation.
    """
    if features_length_fn is None and labels_length_fn is None:
        return lambda dataset: dataset

    def _length_constraints(lengths, maximum_lengths):
        # Work with lists of lengths which correspond to the general multi source case.
        if not isinstance(lengths, list):
            lengths = [lengths]
        if not isinstance(maximum_lengths, list):
            maximum_lengths = [maximum_lengths]
        # Unset maximum lengths are set to None (i.e. no constraint).
        maximum_lengths += [None] * (len(lengths) - len(maximum_lengths))
        constraints = []
        for length, maximum_length in zip(lengths, maximum_lengths):
            constraints.append(tf.greater(length, 0))
            if maximum_length is not None:
                constraints.append(tf.less_equal(length, maximum_length))
        return constraints

    def _predicate(features, labels):
        cond = []
        features_length = (
            features_length_fn(features) if features_length_fn is not None else None
        )
        labels_length = (
            labels_length_fn(labels) if labels_length_fn is not None else None
        )
        if features_length is not None:
            cond.extend(_length_constraints(features_length, maximum_features_length))
        if labels_length is not None:
            cond.extend(_length_constraints(labels_length, maximum_labels_length))
        return tf.reduce_all(cond)

    return lambda dataset: dataset.filter(_predicate)


def make_cardinality_multiple_of(divisor):
    """Transformation that ensures that the dataset cardinality is a multiple of
    :obj:`divisor`.

    Example:

      >>> dataset = tf.data.Dataset.range(7)
      >>> dataset = dataset.apply(opennmt.data.make_cardinality_multiple_of(10))
      >>> len(list(iter(dataset)))
      10

    Args:
      divisor: The value that should divide the dataset size.

    Returns:
      A ``tf.data.Dataset`` transformation.

    Tip:
      This transformation is useful when training multiple replicas on a finite
      dataset. It ensures that each replica receives a non empty batch in the last
      training iteration.
    """
    if divisor == 1:
        return lambda dataset: dataset

    def _continue_iter(num_consumed, element):
        # Continue iterating if the current element is from the original dataset or
        # if the number of consumed batches is not a multiple of divisor.
        is_original = element[0]
        return tf.math.logical_or(
            is_original, tf.math.not_equal(num_consumed % divisor, 0)
        )

    def _retrieve_element(num_consumed, element):
        _ = num_consumed
        return element[1]

    def _transform(dataset):
        # Nothing to do for infinite datasets.
        if dataset.cardinality() == tf.data.INFINITE_CARDINALITY:
            return dataset

        # Concatenate extra batches with a flag.
        extra_batches = dataset.repeat()
        dataset = dataset.map(lambda *x: (tf.constant(True), x))
        extra_batches = extra_batches.map(lambda *x: (tf.constant(False), x))
        dataset = dataset.concatenate(extra_batches)

        # Take all original batches and the number of extra batches required.
        dataset = dataset.enumerate()
        # TODO: clean this API when TensorFlow requirement is updated to >=2.6.
        if compat.tf_supports("data.Dataset.take_while"):
            dataset = dataset.take_while(_continue_iter)
        else:
            dataset = dataset.apply(tf.data.experimental.take_while(_continue_iter))
        return dataset.map(_retrieve_element)  # Retrieve the element only.

    return _transform


def random_shard(shard_size, dataset_size):
    """Transformation that shards the dataset in a random order.

    Example:

      >>> dataset = tf.data.Dataset.range(6)
      >>> dataset = dataset.apply(opennmt.data.random_shard(2, 6))
      >>> list(dataset.as_numpy_iterator())
      [0, 1, 4, 5, 2, 3]

    Args:
      shard_size: The number of examples in each shard.
      dataset_size: The total number of examples in the dataset.

    Returns:
      A ``tf.data.Dataset`` transformation.
    """
    num_shards = -(-dataset_size // shard_size)  # Ceil division.
    offsets = np.linspace(
        0, dataset_size, num=num_shards, endpoint=False, dtype=np.int64
    )

    def _random_shard(dataset):
        sharded_dataset = tf.data.Dataset.from_tensor_slices(offsets)
        sharded_dataset = sharded_dataset.shuffle(num_shards)
        sharded_dataset = sharded_dataset.flat_map(
            lambda offset: dataset.skip(offset).take(shard_size)
        )
        return sharded_dataset

    return _random_shard


def shuffle_dataset(buffer_size, shuffle_shards=True, dataset_size=None):
    """Transformation that shuffles the dataset based on its size.

    Example:

      >>> dataset = tf.data.Dataset.range(6)
      >>> dataset = dataset.apply(opennmt.data.shuffle_dataset(3))
      >>> list(dataset.as_numpy_iterator())
      [2, 3, 1, 0, 4, 5]

    Args:
      buffer_size: The number of elements from which to sample.
      shuffle_shards: When :obj:`buffer_size` is smaller than the dataset size,
        the dataset is first sharded in a random order to add another level of
        shuffling.
      dataset_size: If the dataset size is already known, it can be passed here to
        avoid a slower generic computation of the dataset size later.

    Returns:
      A ``tf.data.Dataset`` transformation.
    """

    def _shuffle(dataset):
        sample_size = buffer_size
        if sample_size < 0 or shuffle_shards:
            total_size = dataset_size
            if total_size is None:
                total_size = get_dataset_size(dataset)
            tf.get_logger().info("Training on %d examples", total_size)
            if sample_size < 0:
                sample_size = total_size
            elif sample_size < total_size:
                dataset = dataset.apply(random_shard(sample_size, total_size))
        dataset = dataset.shuffle(sample_size)
        return dataset

    return _shuffle


def batch_dataset(batch_size, padded_shapes=None):
    """Transformation that batches a dataset.

    Example:

      >>> dataset = dataset.apply(opennmt.data.batch_dataset(...))

    Args:
      batch_size: The batch size.
      padded_shapes: The padded shapes for this dataset. If ``None``, the shapes
        are automatically inferred from the dataset output shapes.

    Returns:
      A ``tf.data.Dataset`` transformation.

    See Also:
      :func:`opennmt.data.batch_sequence_dataset`
    """
    return lambda dataset: dataset.padded_batch(
        batch_size, padded_shapes=padded_shapes or _get_output_shapes(dataset)
    )


def batch_sequence_dataset(
    batch_size,
    batch_type="examples",
    batch_multiplier=1,
    batch_size_multiple=1,
    length_bucket_width=None,
    length_fn=None,
    padded_shapes=None,
):
    """Transformation that batches a dataset of sequences.

    This implements an example-based and a token-based batching strategy
    with optional bucketing of sequences.

    Bucketing makes the batches contain sequences of similar lengths to optimize
    the training efficiency. For example, if :obj:`length_bucket_width` is 5,
    sequences will be organized by the following length buckets:

    1 - 5 | 6 - 10 | 11 - 15 | ...

    Then when building the next batch, sequences will be selected from the same
    length bucket.

    If the dataset has parallel elements (e.g. a parallel source and target
    dataset), the element is assigned to the bucket corresponding to the maximum
    length of all parallel elements.

    Example:

      >>> dataset = dataset.apply(opennmt.data.batch_sequence_dataset(...))

    Args:
      batch_size: The batch size.
      batch_type: The training batching strategy to use: can be "examples" or
        "tokens".
      batch_multiplier: The batch size multiplier.
      batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
        resulting batch size is a multiple of this value.
      length_bucket_width: The width of the length buckets to select batch
        candidates from. ``None`` to not constrain batch formation.
      length_fn: A function or list of functions (in case of a parallel dataset)
        that take features as argument and return the associated sequence length.
      padded_shapes: The padded shapes for this dataset. If ``None``, the shapes
        are automatically inferred from the dataset output shapes.

    Returns:
      A ``tf.data.Dataset`` transformation.

    Raises:
      ValueError: if :obj:`batch_type` is not one of "examples" or "tokens".
      ValueError: if :obj:`batch_type` is "tokens" but :obj:`length_bucket_width`
        is not set.
      ValueError: if the number of length functions in :obj:`length_fn` does not
        match the number of parallel elements.

    See Also:
      :func:`opennmt.data.batch_dataset`
    """
    batch_size = batch_size * batch_multiplier

    def _get_bucket_id(features, length_fn):
        default_id = tf.constant(0, dtype=tf.int32)
        if length_fn is None:
            return default_id
        lengths = length_fn(features)
        if lengths is None:
            return default_id
        if not isinstance(lengths, list):
            lengths = [lengths]  # Fallback to the general case of parallel inputs.
        lengths = [length // length_bucket_width for length in lengths]
        return tf.reduce_max(lengths)

    def _key_func(*args):
        length_fns = length_fn
        if length_fns is None:
            length_fns = [None for _ in args]
        elif not isinstance(length_fns, (list, tuple)):
            length_fns = [length_fns]
        if len(length_fns) != len(args):
            raise ValueError(
                "%d length functions were passed but this dataset contains "
                "%d parallel elements" % (len(length_fns), len(args))
            )
        # Take the highest bucket id.
        bucket_id = tf.reduce_max(
            [
                _get_bucket_id(features, length_fn)
                for features, length_fn in zip(args, length_fns)
            ]
        )
        return tf.cast(bucket_id, tf.int64)

    def _reduce_func(unused_key, dataset):
        return dataset.apply(batch_dataset(batch_size, padded_shapes=padded_shapes))

    def _window_size_func(key):
        if length_bucket_width > 1:
            key += 1  # For length_bucket_width == 1, key 0 is unassigned.
        size = batch_size // (key * length_bucket_width)
        required_multiple = batch_multiplier * batch_size_multiple
        if required_multiple > 1:
            size = size + required_multiple - size % required_multiple
        return tf.cast(tf.maximum(size, required_multiple), tf.int64)

    def _group_by_window(*args, **kwargs):
        # TODO: clean this API when TensorFlow requirement is updated to >=2.6.
        if compat.tf_supports("data.Dataset.group_by_window"):
            return lambda dataset: dataset.group_by_window(*args, **kwargs)
        else:
            return tf.data.experimental.group_by_window(*args, **kwargs)

    if length_bucket_width is None:
        if batch_type == "tokens":
            raise ValueError(
                "Batch type 'tokens' requires length bucketing (the parameter "
                "length_bucket_width should be non null)"
            )
        return batch_dataset(batch_size, padded_shapes=padded_shapes)

    if batch_type == "examples":
        return _group_by_window(_key_func, _reduce_func, window_size=batch_size)
    elif batch_type == "tokens":
        return _group_by_window(
            _key_func, _reduce_func, window_size_func=_window_size_func
        )
    else:
        raise ValueError(
            "Invalid batch type: '{}'; should be 'examples' or 'tokens'".format(
                batch_type
            )
        )


def training_pipeline(
    batch_size,
    batch_type="examples",
    batch_multiplier=1,
    batch_size_multiple=1,
    process_fn=None,
    transform_fns=None,
    length_bucket_width=None,
    features_length_fn=None,
    labels_length_fn=None,
    maximum_features_length=None,
    maximum_labels_length=None,
    single_pass=False,
    num_shards=1,
    shard_index=0,
    num_threads=None,
    dataset_size=None,
    shuffle_buffer_size=None,
    prefetch_buffer_size=None,
    cardinality_multiple=1,
):
    """Transformation that applies most of the dataset operations commonly used
    for training on sequence data:

    * sharding
    * shuffling
    * processing
    * filtering
    * bucketization
    * batching
    * prefetching

    Example:

      >>> dataset = dataset.apply(opennmt.data.training_pipeline(...))

    Args:
      batch_size: The batch size to use.
      batch_type: The training batching strategy to use: can be "examples" or
        "tokens".
      batch_multiplier: The batch size multiplier.
      batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
        resulting batch size is a multiple of this value.
      process_fn: The processing function to apply on each element.
      transform_fns: List of dataset transformation functions (applied after
        :obj:`process_fn` if defined).
      length_bucket_width: The width of the length buckets to select batch
        candidates from. ``None`` to not constrain batch formation.
      features_length_fn: A function mapping features to a sequence length.
      labels_length_fn: A function mapping labels to a sequence length.
      maximum_features_length: The maximum length or list of maximum lengths of
        the features sequence(s). ``None`` to not constrain the length.
      maximum_labels_length: The maximum length of the labels sequence.
        ``None`` to not constrain the length.
      single_pass: If ``True``, makes a single pass over the training data.
      num_shards: The number of data shards (usually the number of workers in a
        distributed setting).
      shard_index: The shard index this data pipeline should read from.
      num_threads: The number of elements processed in parallel.
      dataset_size: If the dataset size is already known, it can be passed here to
        avoid a slower generic computation of the dataset size later.
      shuffle_buffer_size: The number of elements from which to sample.
      prefetch_buffer_size: The number of batches to prefetch asynchronously. If
        ``None``, use an automatically tuned value.
      cardinality_multiple: Ensure that the dataset cardinality is a multiple of
        this value when :obj:`single_pass` is ``True``.

    Returns:
      A ``tf.data.Dataset`` transformation.

    See Also:
      - :func:`opennmt.data.batch_sequence_dataset`
      - :func:`opennmt.data.make_cardinality_multiple_of`
      - :func:`opennmt.data.filter_examples_by_length`
      - :func:`opennmt.data.filter_irregular_batches`
      - :func:`opennmt.data.shuffle_dataset`
    """
    if dataset_size is not None and num_shards > 1:
        # Update dataset_size based on the shard size.
        if isinstance(dataset_size, list):
            dataset_size = [size // num_shards for size in dataset_size]
        else:
            dataset_size //= num_shards

    def _make_weighted_dataset(datasets, weights):
        if single_pass:
            raise ValueError(
                "single_pass parameter is not compatible with weighted datasets"
            )
        if not datasets:
            raise ValueError("At least one dataset is required")
        if weights is not None and len(weights) != len(datasets):
            raise ValueError(
                "%d dataset weights were provided, but %d were expected to match the "
                "number of data files" % (len(weights), len(datasets))
            )
        if num_shards > 1:
            datasets = [dataset.shard(num_shards, shard_index) for dataset in datasets]
        weights = normalize_weights(datasets, weights=weights, sizes=dataset_size)
        datasets = [dataset.repeat() for dataset in datasets]
        # TODO: clean this API when TensorFlow requirement is updated to >=2.7.
        if compat.tf_supports("data.Dataset.sample_from_datasets"):
            dataset = tf.data.Dataset.sample_from_datasets(datasets, weights=weights)
        else:
            dataset = tf.data.experimental.sample_from_datasets(
                datasets, weights=weights
            )
        if shuffle_buffer_size is not None and shuffle_buffer_size != 0:
            if shuffle_buffer_size < 0:
                raise ValueError(
                    "shuffle_buffer_size < 0 is not compatible with weighted datasets"
                )
            dataset = dataset.shuffle(shuffle_buffer_size)
        return dataset

    def _make_single_dataset(dataset):
        if num_shards > 1:
            dataset = dataset.shard(num_shards, shard_index)
        if shuffle_buffer_size is not None and shuffle_buffer_size != 0:
            dataset = dataset.apply(
                shuffle_dataset(shuffle_buffer_size, dataset_size=dataset_size)
            )
        return dataset

    def _pipeline(dataset):
        if isinstance(dataset, tuple):
            dataset, weights = dataset
        else:
            weights = None
        is_weighted_dataset = isinstance(dataset, list)
        if is_weighted_dataset:
            dataset = _make_weighted_dataset(dataset, weights)
        else:
            dataset = _make_single_dataset(dataset)
        if process_fn is not None:
            dataset = dataset.map(process_fn, num_parallel_calls=num_threads or 4)
        if transform_fns is not None:
            for transform_fn in transform_fns:
                dataset = dataset.apply(transform_fn)
        else:
            dataset = dataset.apply(
                filter_examples_by_length(
                    maximum_features_length=maximum_features_length,
                    maximum_labels_length=maximum_labels_length,
                    features_length_fn=features_length_fn,
                    labels_length_fn=labels_length_fn,
                )
            )
        dataset = dataset.apply(
            batch_sequence_dataset(
                batch_size,
                batch_type=batch_type,
                batch_multiplier=batch_multiplier,
                batch_size_multiple=batch_size_multiple,
                length_bucket_width=length_bucket_width,
                length_fn=[features_length_fn, labels_length_fn],
            )
        )
        dataset = dataset.apply(filter_irregular_batches(batch_multiplier))
        if not single_pass:
            if not is_weighted_dataset:  # Weighted dataset is repeated before sampling.
                dataset = dataset.repeat()
        else:
            dataset = dataset.apply(make_cardinality_multiple_of(cardinality_multiple))
        dataset = dataset.prefetch(prefetch_buffer_size)
        return dataset

    return _pipeline


def inference_pipeline(
    batch_size,
    batch_type="examples",
    process_fn=None,
    transform_fns=None,
    length_bucket_width=None,
    length_fn=None,
    num_threads=None,
    prefetch_buffer_size=None,
):
    """Transformation that applies dataset operations for inference.

    Example:

      >>> dataset = dataset.apply(opennmt.data.inference_pipeline(...))

    Args:
      batch_size: The batch size to use.
      batch_type: The batching strategy to use: can be "examples" or "tokens".
      process_fn: The processing function to apply on each element.
      transform_fns: List of dataset transformation functions (applied after
        :obj:`process_fn` if defined).
      length_bucket_width: The width of the length buckets to select batch
        candidates from. If set, this means the inference pipeline will be
        reordered based on the examples length, the application is then
        responsible to restore the predictions in order. An "index" key will be
        inserted in the examples dictionary.
      length_fn: A function mapping features to a sequence length.
      num_threads: The number of elements processed in parallel.
      prefetch_buffer_size: The number of batches to prefetch asynchronously. If
        ``None``, use an automatically tuned value.

    Returns:
      A ``tf.data.Dataset`` transformation.

    Raises:
      ValueError: if :obj:`length_bucket_width` is set but not :obj:`length_fn`.
      ValueError: if :obj:`length_bucket_width` is set but the dataset does not
        output a dictionary structure.
    """

    def _inject_index(index, x):
        if isinstance(x, tuple):
            features = x[0]
        else:
            features = x
        features["index"] = index
        return x

    def _pipeline(dataset):
        if process_fn is not None:
            dataset = dataset.map(process_fn, num_parallel_calls=num_threads)
        if transform_fns is not None:
            for transform_fn in transform_fns:
                dataset = dataset.apply(transform_fn)
        if length_bucket_width is not None and length_bucket_width > 0:
            if length_fn is None:
                raise ValueError("length_fn is required when reordering by length")
            output_shapes = _get_output_shapes(dataset)
            if isinstance(output_shapes, tuple):
                num_length_fn = (
                    len(length_fn) if isinstance(length_fn, (list, tuple)) else 1
                )
                if len(output_shapes) != num_length_fn:
                    raise ValueError(
                        "The dataset outputs %d parallel features, but got %d "
                        "length functions" % (len(output_shapes), num_length_fn)
                    )
                output_shapes = output_shapes[0]
            if not isinstance(output_shapes, dict):
                raise ValueError(
                    "Reordering by length expects dataset elements to be Python dicts"
                )
            dataset = dataset.enumerate()
            dataset = dataset.map(_inject_index)
            dataset = dataset.apply(
                batch_sequence_dataset(
                    batch_size,
                    batch_type=batch_type,
                    length_bucket_width=length_bucket_width,
                    length_fn=length_fn,
                )
            )
        else:
            dataset = dataset.apply(batch_dataset(batch_size))
        dataset = dataset.prefetch(prefetch_buffer_size)
        return dataset

    return _pipeline



"""Checkpoint utilities."""

import copy
import os
import tempfile

import tensorflow as tf

from opennmt.utils import misc


class Checkpoint(object):
    """Wrapper around TensorFlow checkpoints utilities."""

    def __init__(self, model, optimizer=None, model_dir=None, keep_checkpoint_max=8):
        """Initializes the wrapper.

        Args:
          model: A :class:`opennmt.models.Model` to save.
          optimizer: The optimizer instance.
          model_dir: The directory where checkpoints will be saved. If not set, a
            temporary directory will be used.
          keep_checkpoint_max: The maximum number of checkpoints to keep.
        """
        if model_dir is None:
            model_dir = tempfile.mkdtemp()
        trackables = {}
        trackables["model"] = model
        if optimizer is not None:
            trackables["optimizer"] = optimizer
        self._model = model
        self._optimizer = optimizer
        self._model_dir = model_dir
        self._checkpoint = tf.train.Checkpoint(**trackables)
        self._checkpoint_manager = tf.train.CheckpointManager(
            self._checkpoint, model_dir, keep_checkpoint_max
        )

    @classmethod
    def from_config(cls, config, model, optimizer=None):
        """Creates a checkpoint wrapper from the configuration.

        Args:
          config: The user configuration.
          model: A :class:`opennmt.models.Model` to save.
          optimizer: The optimizer instance.

        Returns:
          A :class:`opennmt.utils.Checkpoint` instance.
        """
        train_config = config.get("train")
        if train_config is None:
            train_config = {}
        keep_checkpoint_max = max(
            train_config.get("keep_checkpoint_max", 8),
            train_config.get("average_last_checkpoints", 0),
        )
        return cls(
            model,
            optimizer=optimizer,
            model_dir=config.get("model_dir"),
            keep_checkpoint_max=keep_checkpoint_max,
        )

    @property
    def model(self):
        """The managed model."""
        return self._model

    @property
    def optimizer(self):
        """The managed optimizer."""
        return self._optimizer

    @property
    def model_dir(self):
        """The model directory."""
        return self._model_dir

    @property
    def last_saved_step(self):
        """The last training step that was saved."""
        latest_checkpoint = self._checkpoint_manager.latest_checkpoint
        if latest_checkpoint is None:
            return None
        return get_step_from_checkpoint_prefix(latest_checkpoint)

    def save(self, step=None):
        """Saves a checkpoint.

        Args:
          step: The step to save for. If ``None``, get the value from ``optimizer.iterations``.

        Returns:
          The path to the saved checkpoint.
        """
        if step is None:
            step = self._optimizer.iterations
        path = self._checkpoint_manager.save(checkpoint_number=step)
        tf.get_logger().info("Saved checkpoint %s", path)
        return path

    def restore(self, checkpoint_path=None, weights_only=False):
        """Restores a checkpoint.

        Args:
          checkpoint_path: Path a checkpoint to restore. If not set, the latest
            checkpoint from :obj:`model_dir` will be restored.
          weights_only: Only restore model weights.

        Returns:
          Path to the restored checkpoint.
        """
        if weights_only:
            checkpoint = tf.train.Checkpoint(model=self._model)
        else:
            checkpoint = self._checkpoint
        if checkpoint_path is not None:
            if tf.io.gfile.isdir(checkpoint_path):
                checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)
        elif self._checkpoint_manager.latest_checkpoint is not None:
            checkpoint_path = self._checkpoint_manager.latest_checkpoint
        if checkpoint_path is None:
            tf.get_logger().warning("No checkpoint to restore in %s", self._model_dir)
            return None
        if is_v1_checkpoint(checkpoint_path):
            tf.get_logger().info("Upgrading V1 checkpoint...")
            # Work with copies of model and optimizer as the downstream task might
            # need to create the variable differently (e.g. under a distribution
            # strategy scope).
            tmp_model = misc.clone_layer(self._model)
            tmp_optimizer = (
                copy.deepcopy(self._optimizer) if self._optimizer is not None else None
            )
            tmp_model.create_variables(optimizer=tmp_optimizer)
            step = _restore_v1_checkpoint(
                checkpoint_path, tmp_model, optimizer=tmp_optimizer
            )
            # Save an updated checkpoint in the model directory and restore this one instead.
            tmp_checkpoint = Checkpoint(
                tmp_model, optimizer=tmp_optimizer, model_dir=self._model_dir
            )
            checkpoint_path = tmp_checkpoint.save(step)
            return self.restore(
                checkpoint_path=checkpoint_path, weights_only=weights_only
            )
        load_status = checkpoint.restore(checkpoint_path)
        load_status.expect_partial()
        tf.get_logger().info("Restored checkpoint %s", checkpoint_path)
        return checkpoint_path


def get_step_from_checkpoint_prefix(prefix):
    """Extracts the training step from the checkpoint file prefix."""
    return int(prefix.split("-")[-1])


def is_v1_checkpoint(checkpoint_path):
    """Returns ``True`` if the checkpoint at :obj:`checkpoint_path` has been
    trained with OpenNMT-tf v1.
    """
    if tf.io.gfile.isdir(checkpoint_path):
        checkpoint_path = tf.train.latest_checkpoint(checkpoint_path)
    return os.path.basename(checkpoint_path).startswith("model")


def get_checkpoint_variables(checkpoint_path):
    """Returns variables included in a checkpoint.

    Args:
      checkpoint_path: Path to the checkpoint.

    Returns:
      A dictionary mapping variables name to value.
    """
    reader = tf.train.load_checkpoint(checkpoint_path)
    return {
        name: reader.get_tensor(name)
        for name in reader.get_variable_to_shape_map().keys()
    }


def average_checkpoints(
    model_dir, output_dir, trackables, max_count=8, model_key="model"
):
    """Averages object-based checkpoints.

    Args:
      model_dir: The directory containing checkpoints.
      output_dir: The directory that will contain the averaged checkpoint.
      trackables: A dictionary containing the trackable objects included in the
        checkpoint.
      max_count: The maximum number of checkpoints to average.
      model_key: The key in :obj:`trackables` that references the model.

    Returns:
      The path to the directory containing the averaged checkpoint.

    Raises:
      ValueError: if :obj:`output_dir` is the same as :obj:`model_dir`.
      ValueError: if a model is not found in :obj:`trackables` or is not already
        built.
      ValueError: if no checkpoints are found in :obj:`model_dir`.

    See Also:
      :func:`opennmt.utils.average_checkpoints_into_layer`
    """
    if model_dir == output_dir:
        raise ValueError("Model and output directory must be different")
    model = trackables.get(model_key)
    if model is None:
        raise ValueError("%s not found in trackables %s" % (model_key, trackables))

    checkpoint_state = tf.train.get_checkpoint_state(model_dir)
    if checkpoint_state is None:
        raise ValueError("No checkpoints found in %s" % model_dir)
    checkpoints_path = checkpoint_state.all_model_checkpoint_paths
    if len(checkpoints_path) > max_count:
        checkpoints_path = checkpoints_path[-max_count:]

    average_checkpoints_into_layer(checkpoints_path, model, model_key)

    last_step = get_step_from_checkpoint_prefix(checkpoints_path[-1])
    checkpoint = tf.train.Checkpoint(**trackables)
    new_checkpoint_manager = tf.train.CheckpointManager(
        checkpoint, output_dir, max_to_keep=None
    )
    new_checkpoint_manager.save(checkpoint_number=last_step)
    return output_dir


def average_checkpoints_into_layer(checkpoints, layer, layer_prefix):
    """Updates the layer weights with their average value in the checkpoints.

    Args:
      checkpoints: A non empty list of checkpoint paths.
      layer: A ``tf.keras.layers.Layer`` instance.
      layer_prefix: The name/scope that prefixes the layer variables names in the
        checkpoints.

    Raises:
      ValueError: if :obj:`checkpoints` is empty.
      ValueError: if :obj:`layer` is not already built.

    See Also:
      :func:`opennmt.utils.average_checkpoints`
    """
    if not checkpoints:
        raise ValueError("There should be at least one checkpoint")
    if not layer.built:
        raise ValueError("The layer should be built before calling this function")

    # Reset the layer variables to 0.
    for variable in layer.variables:
        variable.assign(tf.zeros_like(variable))

    # Get a map from variable names in the checkpoint to variables in the layer.
    names_to_variables = misc.get_variables_name_mapping(layer, root_key=layer_prefix)

    num_checkpoints = len(checkpoints)
    tf.get_logger().info("Averaging %d checkpoints...", num_checkpoints)
    for checkpoint_path in checkpoints:
        tf.get_logger().info("Reading checkpoint %s...", checkpoint_path)
        reader = tf.train.load_checkpoint(checkpoint_path)
        for path in reader.get_variable_to_shape_map().keys():
            if not path.startswith(layer_prefix) or ".OPTIMIZER_SLOT" in path:
                continue
            variable = names_to_variables[path]
            value = reader.get_tensor(path)
            variable.assign_add(value / num_checkpoints)


_V1_OPTIM_SCOPE = "optim"
_V1_SLOTS_MAPPING = {"Adam": "m", "Adam_1": "v"}


def _restore_v1_checkpoint(checkpoint_path, model, optimizer=None):
    v1_variables = get_checkpoint_variables(checkpoint_path)
    v1_structure = _variables_to_structure(v1_variables)
    step = v1_structure["global_step"]
    if optimizer is not None:
        optimizer.iterations.assign(step)
        if _V1_OPTIM_SCOPE in v1_structure:
            slots = v1_structure[_V1_OPTIM_SCOPE]
            del v1_structure[_V1_OPTIM_SCOPE]
            v1_structure = _merge_optimizer_slots(v1_structure, slots)
    mapping = model.map_v1_weights(v1_structure)
    existing_variables = set(variable.ref() for variable in model.variables)
    mapped_variables = set(variable.ref() for variable, _ in mapping)
    missing_mapping = existing_variables.difference(mapped_variables)
    if missing_mapping:
        raise ValueError(
            "The following variables were not mapped: %s"
            % (", ".join(var.name for var in missing_mapping))
        )
    # Assign each variable and possibly the optimizer slots.
    for v2_variable, v1_variable in mapping:
        if isinstance(v1_variable, tuple):
            v1_variable, v1_slots = v1_variable
        else:
            v1_slots = None
        v2_variable.assign(v1_variable)
        if v1_slots is not None:
            for slot_name, value in v1_slots.items():
                v2_slot = optimizer.get_slot(v2_variable, slot_name)
                v2_slot.assign(value)
    return step


def _variables_to_structure(variables):
    """Represents variables a nested dictionary with scope names as keys."""
    structure = {}
    for name, value in variables.items():
        fields = name.split("/")
        cur = structure
        for i, key in enumerate(fields):
            if key not in cur:
                if i + 1 == len(fields):
                    cur[key] = value
                    break
                cur[key] = {}
            cur = cur[key]
    return structure


def _merge_optimizer_slots(variables, slots):
    """Replaces leaves in the variables structure by tuples of
    (variable, dict of optimizer slots).
    """
    if isinstance(variables, dict):
        merged = {}
        for key, value in variables.items():
            if key not in slots:
                merged[key] = copy.deepcopy(value)
            else:
                merged[key] = _merge_optimizer_slots(value, slots[key])
        return merged
    else:
        new_slots = {}
        for name, value in slots.items():
            name = _V1_SLOTS_MAPPING.get(name)
            if name is None:
                # Just ignore the optimizer slots if their name is not listed.
                return variables
            new_slots[name] = value
        return (variables, new_slots)
